/*
* -cr_beg-
* ********************************************************************
* IBM Confidential
* OCO Source Materials
* 5724-U38, 5724-U39, 5724-U48, 5655-V17, 5655-V19, 5655-V35, 5655-V37
* (C) Copyright IBM Corp. 2012, 2013
* The source code for this program is not published or otherwise
* divested of its trade secrets, irrespective of what has
* been deposited with the U.S. Copyright Office.
* ********************************************************************
* -cr_end-
*/

//
//  ODPPJNI2103.c
//  odpp
//
//  Created by Peter Hagelund on 8/24/12.
//  Copyright (c) 2012, 2013 IBM. All rights reserved.
//

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <errno.h>
#include <wchar.h>
#include "ODPP.h"
#include "NativeProvider.h"
#include "FrameworkDefinition.h"
#include "ServiceDefinition.h"
#include "InitializationOperand.h"
#include "FieldDefinition.h"
#include "RowSet.h"
#include "Row.h"
#include "Field.h"
#include "ODPPCmnAPI.h"
#include "ODPPCmnErrCodes.h"
#include "ODPPCmnOprH.h"
#include "ODPPCmnIncludesH.h"
#include "ODPPDataTypesH.h"
#include "ODPPPrvDataTypesH.h"

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

/* Just in case we get ahead of the ODPP delivery of 51707 */
#ifndef ODPPSUCCESS_WITH_INFO
#define ODPPSUCCESS_WITH_INFO -2
#endif /* ODPPSUCCESS_WITH_INFO */

/* Synchronization problem with ODPP build/environment. */
#ifndef ENV_APP
#define ENV_APP 'A'
#endif

#define UTF8_MAX_BYTES_PER_CODE_POINT 4
#define CODE_PAGE_UTF8 65001

#define RETURN_IF_NULL(ptr) { if ((void *)(ptr) == (void *)NULL) return; }
#define RETURN_VAL_IF_NULL(ptr,val) { if ((void *)(ptr) == (void *)NULL) return (val); }

// java.lang.String:
static jclass jStringClass = NULL;
// java.lang.Integer:
static jclass jIntegerClass = NULL;
// java.util.ArrayList:
static jclass jArrayListClass = NULL;
static jmethodID jArrayListInitMID = NULL;
static jmethodID jArrayListAddMID = NULL;
static jmethodID jArrayListGetMID = NULL;
static jmethodID jArrayListSizeMID = NULL;
// com.ibm.nex.odpp.ODPPException:
static jclass jODPPExceptionClass = NULL;
static jmethodID jODPPExceptionInitMID = NULL;
// com.ibm.nex.odpp.ODPP:
static jclass jODPPClass = NULL;
static jfieldID jODPPDebugFID = NULL;
// com.ibm.nex.odpp.NativeProvider:
static jclass jNativeProviderClass = NULL;
// com.ibm.nex.odpp.internal.FrameworkDefinition:
static jclass jFrameworkDefinitionClass = NULL;
static jfieldID jFrameworkDefinitionPointerFID = NULL;
static jfieldID jFrameworkDefinitionInitializationOperandsFID = NULL;
// com.ibm.nex.odpp.internal.ServiceDefinition:
static jclass jServiceDefinitionClass = NULL;
static jfieldID jServiceDefinitionPointerFID = NULL;
static jfieldID jServiceDefinitionCopyToDestinationFID = NULL;
static jfieldID jServiceDefinitionInitializationOperandsFID = NULL;
static jfieldID jServiceDefinitionFieldDefinitionsFID = NULL;
static jfieldID jServiceDefinitionServiceIdFID = NULL;
// com.ibm.nex.odpp.internal.InitializationOperand:
static jclass jInitializationOperandClass = NULL;
static jfieldID jInitializationOperandPointerFID = NULL;
static jfieldID jInitializationOperandParameterIdFID = NULL;
static jfieldID jInitializationOperandParameterNameFID = NULL;
static jfieldID jInitializationOperandIValueIdFID = NULL;
static jfieldID jInitializationOperandSValueIdFID = NULL;
static jmethodID jInitializationOperandConstructorMID = NULL;
// com.ibm.nex.odpp.internal.FieldDefinition:
static jclass jFieldDefinitionClass = NULL;
static jfieldID jFieldDefinitionPointerFID = NULL;
static jfieldID jFieldDefinitionIdFID = NULL;
static jfieldID jFieldDefinitionLengthFID = NULL;
static jfieldID jFieldDefinitionPrecisionFID = NULL;
static jfieldID jFieldDefinitionScaleFID = NULL;
static jfieldID jFieldDefinitionColumnNameFID = NULL;
static jmethodID jFieldDefinitionConstructorMID = NULL;
// com.ibm.nex.odpp.RowSet:
static jclass jRowSetClass = NULL;
static jfieldID jRowSetPointerFID = NULL;
static jfieldID jRowSetRowsFID = NULL;
static jfieldID jRowSetHasInfoFID = NULL;
// com.ibm.nex.odpp.Row:
static jclass jRowClass = NULL;
static jfieldID jRowPointerFID = NULL;
static jfieldID jRowHasErrorFID = NULL;
static jfieldID jRowFieldsFID = NULL;
// com.ibm.nex.odpp.Field:
static jclass jFieldClass = NULL;
static jfieldID jFieldPointerFID = NULL;
static jfieldID jFieldIdFID = NULL;
static jfieldID jFieldSourceSizeFID = NULL;
static jfieldID jFieldSourceFID = NULL;
static jfieldID jFieldDestinationSizeFID = NULL;
static jfieldID jFieldDestinationFID = NULL;
static jfieldID jFieldErrorCodeFID = NULL;
static jfieldID jFieldErrorMessageFID = NULL;

static char debug = FALSE;


/* We get away with not synchronizing this in any way, because we are single-threaded when
 * we first call the parser to validate the passed masking strings in the work item.
 */
static unsigned char parserInitialized = FALSE;
static unsigned char parmGenInitialized = FALSE;


/* If wchar is 32-bit convert from Java UTF-16 to platform UTF-32, else use UTF-16 as is. */
static char cvtWchars = (sizeof(ODPP_WCHAR) != 2);

static void checkForError(int svcToken, DP_ROWSET_DEF *pRowsetDef);
static void throwODPPException(JNIEnv *env, const char *message, int code);
static void debugf(const char *format, ...);
static void HexDump(const unsigned char *buffer, int size);
static RETVAL ODPPParseParameters(char *pParamString,  //Provider Syntax string
                           int iParmStrBytes,   //Provider Syntax string bytes
                           int iCodePage,       //Provider Syntax string codepage
                           char cDBMSType,      //Provider Syntax string DBMSType
                           DP_SVC_DEF *pSvcDef, //ODPP required service definition
                                                //structure
                           char *pSvcId,        //Buffer for returning ODPP specific
                                                //service provider name
                           int iSvcIdBytes);

/* Debugging aids for checking structural integrity based on eyecatcher */
#ifdef DEBUG_IOQJNI
#define GENERIC_CHECK_DEF(ptr, eyecatcher, datatype) {\
	    char _eyec[5]; \
	    (void)memcpy(_eyec,(ptr)->cEyeCatcher,4); \
	    _eyec[4]='\0'; \
		if ((ptr) != NULL && strncmp(_eyec, (eyecatcher), sizeof((ptr)->cEyeCatcher)) != 0) {\
			(void)printf("%s at 0x%16.16X fails integrity check. Eyecatcher is %s\n", \
					(datatype),(unsigned long long)(ptr),(ptr)->cEyeCatcher);\
			HexDump((ptr),16);} }

#define CHECK_FRMW_PARAMS_DEF(ptr) GENERIC_CHECK_DEF((ptr), DP_FRMW_PARAMS_DEF_EC, "DP_FRMW_PARAMS_DEF")
#define CHECK_SVC_DEF(ptr) GENERIC_CHECK_DEF((ptr), DP_SVC_DEF_EC, "DP_SVC_DEF")
#define CHECK_INIT_OP_DEF(ptr) GENERIC_CHECK_DEF((ptr), DP_INIT_OP_DEF_EC, "DP_INIT_OP_DEF")
#define CHECK_FIELD_DEF(ptr) GENERIC_CHECK_DEF((ptr), DP_FIELD_DEF_EC, "DP_FIELD_DEF")
#define CHECK_ROWSET_DEF(ptr) GENERIC_CHECK_DEF((ptr), DP_ROWSET_DEF_EC, "DP_ROWSET_DEF")
#define CHECK_ROW_DEF(ptr) GENERIC_CHECK_DEF((ptr), DP_ROW_DEF_EC, "DP_ROW_DEF")
#define CHECK_FIELD_DATA_DEF(ptr) GENERIC_CHECK_DEF(ptr, DP_FIELD_DATA_DEF_EC, "DP_FIELD_DATA_DEF")
#define CHECK_ERR_MSG(ptr) GENERIC_CHECK_DEF((ptr), DP_ERR_MSG_EC, "DP_ERR_MSG")
#define CHECK_ECB(ptr) GENERIC_CHECK_DEF((ptr), DP_ECB_EC, "DP_ECB")
#define PRINT_SIZEOF(x) (void)printf("Size of type " #x " is %d\n", sizeof(x));

#else
#define GENERIC_CHECK_DEF(ptr, eyecatcher, type)
#define CHECK_FRMW_PARAMS_DEF(ptr)
#define CHECK_SVC_DEF(ptr)
#define CHECK_INIT_OP_DEF(ptr)
#define CHECK_FIELD_DEF(ptr)
#define CHECK_ROWSET_DEF(ptr)
#define CHECK_ROW_DEF(ptr)
#define CHECK_FIELD_DATA_DEF(ptr)
#define CHECK_ERR_MSG(ptr)
#define CHECK_ECB(ptr)
#define PRINT_SIZEOF(x)

#endif

/*
 * Class:     com_ibm_nex_odpp_ODPP
 * Method:    init0
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_ODPP_init0(JNIEnv *env, jobject jObject)
{
    jboolean jDebug;
    jclass jClass;
    // java.lang.String:
    jClass = (*env)->FindClass(env, "java/lang/String");
    jStringClass =(*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    // java.lang.Integer:
    jClass = (*env)->FindClass(env, "java/lang/Integer");
    jIntegerClass =(*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    // java.util.ArrayList:
    jClass = (*env)->FindClass(env, "java/util/ArrayList");
    jArrayListClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jArrayListInitMID = (*env)->GetMethodID(env, jArrayListClass, "<init>", "()V");
    jArrayListAddMID = (*env)->GetMethodID(env, jArrayListClass, "add", "(Ljava/lang/Object;)Z");
    jArrayListGetMID = (*env)->GetMethodID(env, jArrayListClass, "get", "(I)Ljava/lang/Object;");
    jArrayListSizeMID = (*env)->GetMethodID(env, jArrayListClass, "size", "()I");
    // com.ibm.nex.odpp.ODPPException:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/ODPPException");
    jODPPExceptionClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jODPPExceptionInitMID = (*env)->GetMethodID(env, jODPPExceptionClass, "<init>", "(Ljava/lang/String;I)V");
    // com.ibm.nex.odpp.ODPP:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/ODPP");
    jODPPClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jODPPDebugFID = (*env)->GetFieldID(env, jODPPClass, "debug", "Z");
    // com.ibm.nex.odpp.NativeProvider:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/NativeProvider");
    jNativeProviderClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    // com.ibm.nex.odpp.internal.FrameworkDefinition:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/internal/FrameworkDefinition");
    jFrameworkDefinitionClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jFrameworkDefinitionPointerFID = (*env)->GetFieldID(env, jFrameworkDefinitionClass, "pointer", "J");
    jFrameworkDefinitionInitializationOperandsFID = (*env)->GetFieldID(env, jFrameworkDefinitionClass, "initializationOperands", "Ljava/util/ArrayList;");
    // com.ibm.nex.odpp.internal.ServiceDefinition:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/internal/ServiceDefinition");
    jServiceDefinitionClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jServiceDefinitionPointerFID = (*env)->GetFieldID(env, jServiceDefinitionClass, "pointer", "J");
    jServiceDefinitionCopyToDestinationFID = (*env)->GetFieldID(env, jServiceDefinitionClass, "copyToDestination", "Z");
    jServiceDefinitionInitializationOperandsFID = (*env)->GetFieldID(env, jServiceDefinitionClass, "initializationOperands", "Ljava/util/ArrayList;");
    jServiceDefinitionFieldDefinitionsFID = (*env)->GetFieldID(env, jServiceDefinitionClass, "fieldDefinitions", "Ljava/util/ArrayList;");
    jServiceDefinitionServiceIdFID = (*env)->GetFieldID(env, jServiceDefinitionClass, "serviceId", "Ljava/lang/String;");
    // com.ibm.nex.odpp.internal.InitializationOperand:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/internal/InitializationOperand");
    jInitializationOperandClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jInitializationOperandPointerFID = (*env)->GetFieldID(env, jInitializationOperandClass, "pointer", "J");
    jInitializationOperandParameterIdFID = (*env)->GetFieldID(env, jInitializationOperandClass, "parameterId", "I");
    jInitializationOperandParameterNameFID = (*env)->GetFieldID(env, jInitializationOperandClass, "parameterName", "Ljava/lang/String;");
    jInitializationOperandIValueIdFID = (*env)->GetFieldID(env, jInitializationOperandClass, "iValue", "I");
    jInitializationOperandSValueIdFID = (*env)->GetFieldID(env, jInitializationOperandClass, "sValue", "Ljava/lang/String;");
    jInitializationOperandConstructorMID = (*env)->GetMethodID(env, jInitializationOperandClass, "<init>", "()V");
    // com.ibm.nex.odpp.internal.FieldDefinition:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/internal/FieldDefinition");
    jFieldDefinitionClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jFieldDefinitionPointerFID = (*env)->GetFieldID(env, jFieldDefinitionClass, "pointer", "J");
    jFieldDefinitionIdFID = (*env)->GetFieldID(env, jFieldDefinitionClass, "id", "I");
    jFieldDefinitionLengthFID = (*env)->GetFieldID(env, jFieldDefinitionClass, "length", "I");
    jFieldDefinitionPrecisionFID = (*env)->GetFieldID(env, jFieldDefinitionClass, "precision", "I");
    jFieldDefinitionScaleFID = (*env)->GetFieldID(env, jFieldDefinitionClass, "scale", "I");
    jFieldDefinitionColumnNameFID = (*env)->GetFieldID(env, jFieldDefinitionClass, "columnName", "Ljava/lang/String;");
    jFieldDefinitionConstructorMID = (*env)->GetMethodID(env, jFieldDefinitionClass, "<init>", "()V");
    // com.ibm.nex.odpp.RowSet:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/RowSet");
    jRowSetClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jRowSetPointerFID = (*env)->GetFieldID(env, jRowSetClass, "pointer", "J");
    jRowSetRowsFID = (*env)->GetFieldID(env, jRowSetClass, "rows", "[Lcom/ibm/nex/odpp/Row;");
    jRowSetHasInfoFID = (*env)->GetFieldID(env, jRowSetClass, "rowSetHasInfo", "Z");
    // com.ibm.nex.odpp.Row:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/Row");
    jRowClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jRowPointerFID = (*env)->GetFieldID(env, jRowClass, "pointer", "J");
    jRowHasErrorFID = (*env)->GetFieldID(env, jRowClass, "hasError", "Z");
    jRowFieldsFID = (*env)->GetFieldID(env, jRowClass, "fields", "[Lcom/ibm/nex/odpp/Field;");
    // com.ibm.nex.odpp.Field:
    jClass = (*env)->FindClass(env, "com/ibm/nex/odpp/Field");
    jFieldClass = (*env)->NewGlobalRef(env, jClass);
    (*env)->DeleteLocalRef(env, jClass);
    jFieldPointerFID = (*env)->GetFieldID(env, jFieldClass, "pointer", "J");
    jFieldIdFID = (*env)->GetFieldID(env, jFieldClass, "id", "I");
    jFieldSourceSizeFID = (*env)->GetFieldID(env, jFieldClass, "sourceSize", "I");
    jFieldSourceFID = (*env)->GetFieldID(env, jFieldClass, "source", "Ljava/lang/Object;");
    jFieldDestinationSizeFID = (*env)->GetFieldID(env, jFieldClass, "destinationSize", "I");
    jFieldDestinationFID = (*env)->GetFieldID(env, jFieldClass, "destination", "Ljava/lang/Object;");
    jFieldErrorCodeFID = (*env)->GetFieldID(env, jFieldClass, "errorCode", "I");
    jFieldErrorMessageFID = (*env)->GetFieldID(env, jFieldClass, "errorMessage", "Ljava/lang/String;");
    // Set up debugging
    jDebug = (*env)->GetBooleanField(env, jObject, jODPPDebugFID);
    debug = jDebug;
	PRINT_SIZEOF(DP_FRMW_PARAMS_DEF)
	PRINT_SIZEOF(DP_SVC_DEF)
	PRINT_SIZEOF(DP_FIELD_DEF)
	PRINT_SIZEOF(DP_ROWSET_DEF)
	PRINT_SIZEOF(DP_ROW_DEF)
	PRINT_SIZEOF(DP_FIELD_DATA_DEF)
}

/*
 * Class:     com_ibm_nex_odpp_ODPP
 * Method:    parseODPPSyntaxString
 * Signature: (Ljava/lang/String;Lcom/ibm/nex/odpp/internal/ServiceDefinition;)V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_ODPP_parseODPPSyntaxString(JNIEnv *env, jobject jObject, jstring jMaskDefinition, jobject jServiceDefinition)
{
	RETVAL retval;
	char *odppMaskingString = (char *)NULL;
	int  numMaskingStringBytes = 0;
	int  maskingStringBufferBytes = 0;
	DP_SVC_DEF *pSvcDef = (DP_SVC_DEF *)NULL;
	DP_SVC_DEF *pTmpSvcDef;
	char svcId[ODPP_PROVIDER_ID_MAX_LENGTH];
    const jchar *jMaskDefinitionChars;
    jsize jMaskDefinitionLength = 0;
    jlong jServiceDefinitionPointer = 0L;
    jsize jSvcIdLength = 0;
    jchar *jChars = (jchar *)NULL;
    int jSvcIdSize = 0;
    jobject jSource;
    jlong jPointer;
    jchar tmpBuf[256];

    RETURN_IF_NULL(jServiceDefinitionPointerFID);
    debugf("parseODPPSyntaxString: Getting SyntaxString...\n");
    jMaskDefinitionLength = (*env)->GetStringLength(env, jMaskDefinition);
    debugf("parseODPPSyntaxString: Mask definitions length is %d.\n", jMaskDefinitionLength);

    maskingStringBufferBytes = (jMaskDefinitionLength * UTF8_MAX_BYTES_PER_CODE_POINT);
    if (maskingStringBufferBytes+1 <= sizeof(tmpBuf)){
    	odppMaskingString = (char *)tmpBuf;
    } else {
    	odppMaskingString = (char *)malloc(maskingStringBufferBytes + 1);
    }
    if (odppMaskingString == (char *)NULL) {
        /* debugf("parseODPPSyntaxString: malloc of masking string failed \n"); // If malloc failed, the allocation for the message will, too */
        throwODPPException(env, "parseODPPSyntaxString malloc failed", ODPPFAILURE);
        return;
    }
    memset(odppMaskingString, 0, maskingStringBufferBytes + 1);
    debugf("parseODPPSyntaxString: allocated buffer, getting string chars\n");

    jMaskDefinitionChars = (*env)->GetStringChars(env, jMaskDefinition, NULL);
    debugf("parseODPPSyntaxString: got String chars, converting to 8 bit\n");
    numMaskingStringBytes = 0;
    Convert16To8((unsigned char *)odppMaskingString, maskingStringBufferBytes, jMaskDefinitionChars, jMaskDefinitionLength, &numMaskingStringBytes);
    debugf("parseODPPSyntaxString: 8 bit parse string is \"%s\"\n", odppMaskingString);
    (*env)->ReleaseStringChars(env, jMaskDefinition, jMaskDefinitionChars);
    jMaskDefinitionChars = NULL;

    /* Allocate DP_SVC_DEF */
    debugf("parseODPPSyntaxString: Allocating 'DP_SVC_DEF'...\n");
    pSvcDef = (DP_SVC_DEF *)malloc(sizeof(DP_SVC_DEF));
    memset(pSvcDef, 0, sizeof(DP_SVC_DEF));
    INITIALIZE_ODPP_STRUCT_PTR(pSvcDef, DP_SVC_DEF)

    debugf("parseODPPSyntaxString: Calling ODPPParseParameters.\n");

	//retval = ODPPParseParameters(odppMaskingString, numMaskingStringBytes, CODE_PAGE_UTF8, RDB_NONE, pSvcDef, svcId, sizeof(svcId));
	retval = ODPPParseParameters(odppMaskingString, STRLEN(odppMaskingString), CODE_PAGE_UTF8, RDB_NONE, pSvcDef, svcId, sizeof(svcId));
	if (odppMaskingString != NULL && odppMaskingString != (char *)tmpBuf) {
		free(odppMaskingString);
	}
	odppMaskingString = NULL;
	if (retval != ODPPSUCCESS) {
        debugf("parseODPPSyntaxString: Call to ODPPParseParameters failed \n");
        throwODPPException(env, "parseODPPSyntaxString ODPPParseParameters failed", retval);
        return;
	}
	/* Service ID needs to go SOMEWHERE, and pSvcDef needs to be attached to the ServiceDefinition */
    debugf("parseODPPSyntaxString: Service id is '%s'.\n", svcId);
    jServiceDefinitionPointer = (*env)->GetLongField(env, jServiceDefinition, jServiceDefinitionPointerFID);
    debugf("parseODPPSyntaxString: The ServiceDefinition pointer is 0x%16.16X.\n", jServiceDefinitionPointer);
    if (0L != jServiceDefinitionPointer) {
    	pTmpSvcDef = (DP_SVC_DEF *) jServiceDefinitionPointer;
    	free(pTmpSvcDef); /* NUKE IT.  But really, it shouldn't be anything but NULL */
    	pTmpSvcDef = NULL;
    }
    CHECK_SVC_DEF(pSvcDef);
    jPointer = (jlong) pSvcDef;
    (*env)->SetLongField(env, jServiceDefinition, jServiceDefinitionPointerFID, jPointer);
    debugf("parseODPPSyntaxString: ServiceDefinition Pointer is 0x%16.16X and set on ServiceDefinition Object\n", jPointer);

    /* Convert svcId to a 16-bit Unicode string and populate */
    jSvcIdLength = STRLEN(svcId);
    debugf("parseODPPSyntaxString: Source string length is %d.\n", jSvcIdLength);
    if ((jSvcIdLength + 1) * sizeof(jchar) <= sizeof(tmpBuf)) {
    	jChars = tmpBuf;
    } else {
    	jChars = (jchar *)malloc(sizeof(jchar) * (jSvcIdLength + 1));
    }
    memset(jChars, 0, sizeof(jchar) * (jSvcIdLength + 1));
  	Convert8To16(jChars, jSvcIdLength, (unsigned char *)svcId, jSvcIdLength, &jSvcIdSize);
  	debugf("Converted 8 to 16. jSvcIdSize = %d, jSvcIdLength = %d\n", jSvcIdSize, jSvcIdLength);
  	jSource = (*env)->NewString(env, jChars, jSvcIdLength);
    (*env)->SetObjectField(env, jServiceDefinition, jServiceDefinitionServiceIdFID, jSource);
   	if (jChars != NULL && jChars != tmpBuf) {
   		free(jChars);
   	}
   	jChars = NULL;
   	(*env)->DeleteLocalRef(env, jSource);
}

/*
 * Class:     com_ibm_nex_odpp_ODPP
 * Method:    initializeFramework0
 * Signature: (Lcom/ibm/nex/odpp/internal/FrameworkDefinition;)V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_ODPP_initializeFramework0(JNIEnv *env, jobject jObject, jobject jFrameworkDefinition)
{
    jlong jFrameworkDefinitionPointer;
    RETVAL retVal;
    DP_FRMW_PARAMS_DEF *pDpFrmwParamsDef;
    debugf("initializeFramework0: Initializing framework...\n");
    pDpFrmwParamsDef = NULL;
    if (jFrameworkDefinition != NULL) {
        debugf("initializeFramework0: A 'FrameworkDefinition' has been supplied...\n");
        jFrameworkDefinitionPointer = (*env)->GetLongField(env, jFrameworkDefinition, jFrameworkDefinitionPointerFID);
        debugf("initializeFramework0: The pointer is 0x%16.16X.\n", jFrameworkDefinitionPointer);
        pDpFrmwParamsDef = (DP_FRMW_PARAMS_DEF *)jFrameworkDefinitionPointer;
    } else {
        debugf("initializeFramework0: A 'FrameworkDefinition' not not been supplied - using defaults...\n");
    }
    debugf("initializeFramework0: Invoking 'Provider_FrmwInit'...\n");
    retVal = Provider_FrmwInit(pDpFrmwParamsDef);
    debugf("initializeFramework0: Result is %d\n", retVal);
    if (retVal != ODPPSUCCESS) {
        debugf("initializeFramework0: Signaling framework initialization failure...\n");
        throwODPPException(env, "Provider_FrmwInit failed", retVal);
        return;
    }
    debugf("initializeFramework0: Framework initialized.\n");
}

/*
 * Class:     com_ibm_nex_odpp_ODPP
 * Method:    terminateFramework0
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_ODPP_terminateFramework0(JNIEnv *env, jobject jObject)
{
    RETVAL retVal;
    debugf("terminateFramework0: Terminating framework...\n");
    debugf("terminateFramework0: Invoking 'Provider_FrmwTerm'...\n");
    retVal = Provider_FrmwTerm();
    debugf("terminateFramework0: Result is %d\n", retVal);
    if (retVal != ODPPSUCCESS) {
        debugf("terminateFramework0: Signaling framework termination failure...\n");
        throwODPPException(env, "Provider_FrmwTerm failed", retVal);
        return;
    }
    debugf("terminateFramework0: Framework terminated.\n");
}

/*
 * Class:     com_ibm_nex_odpp_ODPP
 * Method:    destroy0
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_ODPP_destroy0(JNIEnv *env, jobject jObject)
{
    (*env)->DeleteGlobalRef(env, jFieldClass);
    (*env)->DeleteGlobalRef(env, jFieldDefinitionClass);
    (*env)->DeleteGlobalRef(env, jInitializationOperandClass);
    (*env)->DeleteGlobalRef(env, jServiceDefinitionClass);
    (*env)->DeleteGlobalRef(env, jFrameworkDefinitionClass);
    (*env)->DeleteGlobalRef(env, jNativeProviderClass);
    (*env)->DeleteGlobalRef(env, jODPPClass);
    (*env)->DeleteGlobalRef(env, jODPPExceptionClass);
    (*env)->DeleteGlobalRef(env, jArrayListClass);
    (*env)->DeleteGlobalRef(env, jIntegerClass);
    (*env)->DeleteGlobalRef(env, jStringClass);
}

/*
 * Class:     com_ibm_nex_odpp_NativeProvider
 * Method:    initializeProvider0
 * Signature: (Ljava/lang/String;Lcom/ibm/nex/odpp/internal/ServiceDefinition;Z)I
 */
JNIEXPORT jint JNICALL Java_com_ibm_nex_odpp_NativeProvider_initializeProvider0(JNIEnv *env, jobject jObject, jstring jServiceId, jobject jServiceDefinition, jboolean jAllocateInitBlock)
{
    jsize jServiceIdLength;
    const jchar *jServiceIdChars;
    char serviceId[256];
    int jServiceIdSize;
    jlong jServiceDefinitionPointer;
    DP_SVC_DEF *pDpSvcDef;
    char allocateInitBlock;
    int iSvcToken;
    RETVAL retVal;
    RETURN_VAL_IF_NULL(jServiceDefinitionPointerFID,0);
    debugf("initializeProvider0: Initializing provider...\n");
    jServiceIdLength = (*env)->GetStringLength(env, jServiceId);
    debugf("initializeProvider0: Service id length is %d.\n", jServiceIdLength);
    memset(serviceId, 0, sizeof(serviceId));
    jServiceIdChars = (*env)->GetStringChars(env, jServiceId, NULL);
    jServiceIdSize = 0;
    Convert16To8((unsigned char *)serviceId, sizeof(serviceId), jServiceIdChars, jServiceIdLength, &jServiceIdSize);
    (*env)->ReleaseStringChars(env, jServiceId, jServiceIdChars);
    jServiceIdChars = NULL;
    debugf("initializeProvider0: Service id is '%s'.\n", serviceId);
    jServiceDefinitionPointer = (*env)->GetLongField(env, jServiceDefinition, jServiceDefinitionPointerFID);
    debugf("initializeProvider0: The pointer is 0x%16.16X.\n", jServiceDefinitionPointer);
    pDpSvcDef = (DP_SVC_DEF *)jServiceDefinitionPointer;
    CHECK_SVC_DEF(pDpSvcDef);
    allocateInitBlock = (char) jAllocateInitBlock;
    debugf("initializeProvider0: Invoking 'Provider_Init'...\n");
    iSvcToken = 0;
    retVal = Provider_Init(&iSvcToken, serviceId, jServiceIdLength, pDpSvcDef, allocateInitBlock);
    debugf("initializeProvider0: Result is %d\n", retVal);
    if (retVal != ODPPSUCCESS) {
        debugf("initializeProvider0: Signaling provider initialization failure...\n");
        throwODPPException(env, "Provider_Init failed", retVal);
        return 0;
    }
    debugf("initializeProvider0: Provider initialized.\n");
    return iSvcToken;
}

/*
 * Class:     com_ibm_nex_odpp_NativeProvider
 * Method:    serviceProvider0
 * Signature: (ILcom/ibm/nex/odpp/internal/ServiceDefinition;Lcom/ibm/nex/odpp/RowSet;)V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_NativeProvider_serviceProvider0(JNIEnv *env, jobject jObject, jint jServiceToken, jobject jServiceDefinition, jobject jRowSet)
{
    int iSvcToken;
    jboolean jHasInfo;
    jlong jServiceDefinitionPointer;
    DP_SVC_DEF *pDpSvcDef;
    jlong jRowSetPointer;
    DP_ROWSET_DEF *pDpRowsetDef;
    DP_ROW_DEF *pDpRowDefine;
    RETVAL retVal;
    RETURN_IF_NULL(jServiceDefinitionPointerFID);
    debugf("serviceProvider0: Servicing provider...\n");
    iSvcToken = jServiceToken;
    jServiceDefinitionPointer = (*env)->GetLongField(env, jServiceDefinition, jServiceDefinitionPointerFID);
    pDpSvcDef = (DP_SVC_DEF *)jServiceDefinitionPointer;
    CHECK_SVC_DEF(pDpSvcDef);
    /* We are expecting pDpSvcDef to be NULL, so don't fret if it is */
    jRowSetPointer = (*env)->GetLongField(env, jRowSet, jRowSetPointerFID);
    pDpRowsetDef = (DP_ROWSET_DEF *)jRowSetPointer;
    if (pDpRowsetDef == (DP_ROWSET_DEF *)NULL) {
    	/* If we have a null pointer here, we have a problem */
    	debugf("serviceProvider0: pDpRowsetDef is NULL. Problem.");
        throwODPPException(env, "pDpRowSetDef is NULL", ODPPFAILURE);
        return;
    }
    CHECK_ROWSET_DEF(pDpRowsetDef);
    pDpRowDefine = pDpRowsetDef->pRowDefine;
    /* If we don't have a row in the rowset, there is a problem */
    if (pDpRowDefine == (DP_ROW_DEF *)NULL) {
    	/* If we have a null pointer here, we have a problem */
    	debugf("serviceProvider0: pDpRowsetDef->pRowDefine is NULL. There should be at least one row.");
        throwODPPException(env, "pDpRowDefine is NULL", ODPPFAILURE);
        return;
    }
    // This is necessary in order to force ODPP to process all fields in all rows over and
    while (pDpRowDefine != NULL) {
        CHECK_ROW_DEF(pDpRowDefine);
        pDpRowDefine->bProcessed = FALSE;
        pDpRowDefine = pDpRowDefine->pNext;
    }
    debugf("serviceProvider0: Invoking 'Provider_Service'...\n");
    retVal = Provider_Service(iSvcToken, ODPP_METHOD_DEFAULT, pDpSvcDef, pDpRowsetDef);
    debugf("Result is %d\n", retVal);
    if (retVal != ODPPSUCCESS && retVal != ODPPSUCCESS_WITH_INFO) {
        debugf("serviceProvider0: Signaling provider service failure...\n");
        throwODPPException(env, "Provider_Service failed", retVal);
        return;
    }
    if (retVal == ODPPSUCCESS_WITH_INFO) {
    	checkForError(iSvcToken, pDpRowsetDef);
    	/* Set the flag on the RowSet */
        jHasInfo = TRUE;
        (*env)->SetBooleanField(env, jRowSet, jRowSetHasInfoFID, jHasInfo);
    }
    debugf("serviceProvider0: Provider serviced.\n");
}

/*
 * Class:     com_ibm_nex_odpp_NativeProvider
 * Method:    terminateProvider0
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_NativeProvider_terminateProvider0(JNIEnv *env, jobject jObject, jint jServiceToken)
{
    int iSvcToken;
    RETVAL retVal;
    debugf("terminateProvider0: Terminating provider...\n");
    iSvcToken = jServiceToken;
    debugf("terminateProvider0: Invoking 'Provider_Term'...\n");
    retVal = Provider_Term(iSvcToken);
    debugf("terminateProvider0: Result is %d\n", retVal);
    if (retVal != ODPPSUCCESS) {
        debugf("terminateProvider0: Signaling provider termination failure...\n");
        throwODPPException(env, "Provider_Term failed", retVal);
        return;
    }
    debugf("terminateProvider0: Provider terminated.\n");
}

/*
 * Class:     com_ibm_nex_odpp_internal_FrameworkDefinition
 * Method:    toC
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_FrameworkDefinition_toC(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_FRMW_PARAMS_DEF *pDpFrmwParamsDef;
    jobject jInitializationOperands;
    jint jInitializationOperandsSize;
    DP_INIT_OP_DEF *pCurrent;
    int i;
    jobject jInitializationOperand;
    jlong jInitializationOperandPointer;
    DP_INIT_OP_DEF *pDpInitOpDef;
    RETURN_IF_NULL(jFrameworkDefinitionPointerFID);
    debugf("FrameworkDefinition_toC: Populating 'DP_FRMW_PARAMS_DEF' from 'FrameworkDefinition'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jFrameworkDefinitionPointerFID);
    debugf("FrameworkDefinition_toC: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        debugf("FrameworkDefinition_toC: Allocating 'FrameworkDefinition'...\n");
        debugf("FrameworkDefinition_toC: Allocating 'DP_FRMW_PARAMS_DEF'...\n");
        pDpFrmwParamsDef = (DP_FRMW_PARAMS_DEF *)malloc(sizeof(DP_FRMW_PARAMS_DEF));
        memset(pDpFrmwParamsDef, 0, sizeof(DP_FRMW_PARAMS_DEF));
        INITIALIZE_ODPP_STRUCT_PTR(pDpFrmwParamsDef, DP_FRMW_PARAMS_DEF)
        jPointer = (jlong) pDpFrmwParamsDef;
        (*env)->SetLongField(env, jObject, jFrameworkDefinitionPointerFID, jPointer);
        debugf("FrameworkDefinition_toC: Allocation of 'FrameworkDefinition' done. Pointer is 0x%16.16X\n", jPointer);
    }
    pDpFrmwParamsDef = (DP_FRMW_PARAMS_DEF *)jPointer;
    CHECK_FRMW_PARAMS_DEF(pDpFrmwParamsDef)
    jInitializationOperands = (*env)->GetObjectField(env, jObject, jFrameworkDefinitionInitializationOperandsFID);
    jInitializationOperandsSize = (*env)->CallIntMethod(env, jInitializationOperands, jArrayListSizeMID);
    debugf("FrameworkDefinition_toC: Initialization operands size is %d.\n", jInitializationOperandsSize);
    pDpFrmwParamsDef->sParamCount = (short)jInitializationOperandsSize;
    if (jInitializationOperandsSize > 0) {
        debugf("FrameworkDefinition_toC: Linking initialization operands...\n");
        pCurrent = NULL;
        for (i = 0; i < jInitializationOperandsSize; i++) {
            jInitializationOperand = (*env)->CallObjectMethod(env, jInitializationOperands, jArrayListGetMID, i);
            jInitializationOperandPointer = (*env)->GetLongField(env, jInitializationOperand, jInitializationOperandPointerFID);
            pDpInitOpDef = (DP_INIT_OP_DEF *)jInitializationOperandPointer;
            CHECK_INIT_OP_DEF(pDpInitOpDef)
            if (pCurrent == NULL) {
                pDpFrmwParamsDef->pParams = pDpInitOpDef;
            } else {
                pCurrent->pNext = pDpInitOpDef;
            }
            pCurrent = pDpInitOpDef;
            (*env)->DeleteLocalRef(env, jInitializationOperand);
        }
        pCurrent->pNext = (DP_INIT_OP_DEF *)NULL;
    }
    debugf("FrameworkDefinition_toC:  Done\n");

}

/*
 * Class:     com_ibm_nex_odpp_internal_FrameworkDefinition
 * Method:    toJava
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_FrameworkDefinition_toJava(JNIEnv *env, jobject jObject)
{
    // Do nothing.
}

/*
 * Class:     com_ibm_nex_odpp_internal_FrameworkDefinition
 * Method:    free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_FrameworkDefinition_free(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_FRMW_PARAMS_DEF *pDpFrmwParamsDef;
    RETURN_IF_NULL(jFrameworkDefinitionPointerFID);
    debugf("FrameworkDefinition_free: Freeing 'FrameworkDefinition'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jFrameworkDefinitionPointerFID);
    debugf("FrameworkDefinition_free: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer != 0L) {
        pDpFrmwParamsDef = (DP_FRMW_PARAMS_DEF *)jPointer;
        CHECK_FRMW_PARAMS_DEF(pDpFrmwParamsDef)
        debugf("FrameworkDefinition_free: DP_FRMW_PARAMS_DEF is 0x%16.16X\n", (jlong) pDpFrmwParamsDef);
        debugf("FrameworkDefinition_free: Freeing 'DP_FRMW_PARAMS_DEF'...\n");
        free(pDpFrmwParamsDef);
        jPointer = 0L;
        (*env)->SetLongField(env, jObject, jFrameworkDefinitionPointerFID, jPointer);
        debugf("FrameworkDefinition_free: Free of 'FrameworkDefinition' done.\n");
    }
    debugf("FrameworkDefinition_free: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_internal_ServiceDefinition
 * Method:    toC
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_ServiceDefinition_toC(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_SVC_DEF *pDpSvcDef;
    jboolean jCopyToDestination;
    jobject jInitializationOperands;
    jint jInitializationOperandsSize;
    DP_INIT_OP_DEF *pCurrentDpInitOpDef;
    int i;
    jobject jInitializationOperand;
    jlong jInitializationOperandPointer;
    DP_INIT_OP_DEF *pDpInitOpDef;
    jobject jFieldDefinitions;
    jint jFieldDefinitionsSize;
    DP_FIELD_DEF *pCurrentDpFieldDef;
    jobject jFieldDefinition;
    jlong jFieldDefinitionPointer;
    DP_FIELD_DEF *pDpFieldDef;
    RETURN_IF_NULL(jServiceDefinitionPointerFID);
    debugf("ServiceDefinition_toC: Populating 'DP_SVC_DEF' from 'ServiceDefinition'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jServiceDefinitionPointerFID);
    debugf("ServiceDefinition_toC: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        debugf("ServiceDefinition_toC: Allocating 'ServiceDefinition'...\n");
        debugf("ServiceDefinition_toC: Allocating 'DP_SVC_DEF'...\n");
        pDpSvcDef = (DP_SVC_DEF *)malloc(sizeof(DP_SVC_DEF));
        memset(pDpSvcDef, 0, sizeof(DP_SVC_DEF));
        INITIALIZE_ODPP_STRUCT_PTR(pDpSvcDef, DP_SVC_DEF)
        jPointer = (jlong) pDpSvcDef;
        (*env)->SetLongField(env, jObject, jServiceDefinitionPointerFID, jPointer);
        debugf("ServiceDefinition_toC: Allocation of 'ServiceDefinition' done. Pointer is 0x%16.16X\n", jPointer);
    }
    pDpSvcDef = (DP_SVC_DEF *)jPointer;
    CHECK_SVC_DEF(pDpSvcDef)
    jCopyToDestination = (*env)->GetBooleanField(env, jObject, jServiceDefinitionCopyToDestinationFID);
    pDpSvcDef->bCopyToDest = jCopyToDestination;
    debugf("ServiceDefinition_toC: Copy to destination is %s\n", (pDpSvcDef->bCopyToDest ? "true" : "false"));
    jInitializationOperands = (*env)->GetObjectField(env, jObject, jServiceDefinitionInitializationOperandsFID);
    jInitializationOperandsSize = (*env)->CallIntMethod(env, jInitializationOperands, jArrayListSizeMID);
    debugf("ServiceDefinition_toC: Initialization operands size is %d.\n", jInitializationOperandsSize);
    pDpSvcDef->sParamCount = (short)jInitializationOperandsSize;
    if (jInitializationOperandsSize > 0) {
        debugf("ServiceDefinition_toC: Linking initialization operands...\n");
        pCurrentDpInitOpDef = NULL;
        for (i = 0; i < jInitializationOperandsSize; i++) {
            jInitializationOperand = (*env)->CallObjectMethod(env, jInitializationOperands, jArrayListGetMID, i);
            jInitializationOperandPointer = (*env)->GetLongField(env, jInitializationOperand, jInitializationOperandPointerFID);
            pDpInitOpDef = (DP_INIT_OP_DEF *)jInitializationOperandPointer;
            CHECK_INIT_OP_DEF(pDpInitOpDef)
            if (pCurrentDpInitOpDef == NULL) {
                pDpSvcDef->pParams = pDpInitOpDef;
            } else {
                pCurrentDpInitOpDef->pNext = pDpInitOpDef;
            }
            pCurrentDpInitOpDef = pDpInitOpDef;
            (*env)->DeleteLocalRef(env, jInitializationOperand);
        }
        pCurrentDpInitOpDef->pNext = (DP_INIT_OP_DEF *)NULL;
    }
    jFieldDefinitions = (*env)->GetObjectField(env, jObject, jServiceDefinitionFieldDefinitionsFID);
    jFieldDefinitionsSize = (*env)->CallIntMethod(env, jFieldDefinitions, jArrayListSizeMID);
    debugf("ServiceDefinition_toC: Field definitions size is %d.\n", jFieldDefinitionsSize);
    pDpSvcDef->sFldCount = (short)jFieldDefinitionsSize;
    if (jFieldDefinitionsSize > 0) {
        debugf("ServiceDefinition_toC: Linking field definitions...\n");
        pCurrentDpFieldDef = NULL;
        for (i = 0; i < jFieldDefinitionsSize; i++) {
            jFieldDefinition = (*env)->CallObjectMethod(env, jFieldDefinitions, jArrayListGetMID, i);
            jFieldDefinitionPointer = (*env)->GetLongField(env, jFieldDefinition, jFieldDefinitionPointerFID);
            pDpFieldDef = (DP_FIELD_DEF *)jFieldDefinitionPointer;
            CHECK_FIELD_DEF(pDpFieldDef)
            if (pCurrentDpFieldDef == NULL) {
                pDpSvcDef->pFldDef = pDpFieldDef;
            } else {
                pCurrentDpFieldDef->pNext = pDpFieldDef;
            }
            pCurrentDpFieldDef = pDpFieldDef;
            (*env)->DeleteLocalRef(env, jFieldDefinition);
        }
        pCurrentDpFieldDef->pNext = (DP_FIELD_DEF *)NULL;
    }
    debugf("ServiceDefinition_toC:Done.\n");

}

/*
 * Class:     com_ibm_nex_odpp_internal_ServiceDefinition
 * Method:    toJava
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_ServiceDefinition_toJava(JNIEnv *env, jobject jObject)
{
	jlong jPointer;
	DP_SVC_DEF *pDpSvcDef;
	jboolean jCopyToDestination;
	jboolean addSucceeded;
	jobject jInitializationOperands;
	jint jInitializationOperandsSize;
	jobject jInitializationOperand;
	jlong jInitializationOperandPointer;
	DP_INIT_OP_DEF *pDpInitOpDef;
	jobject jFieldDefinitions;
	jint jFieldDefinitionsSize;
	jobject jFieldDefinition;
	jlong jFieldDefinitionPointer;
	DP_FIELD_DEF *pDpFieldDef;

	RETURN_IF_NULL(jServiceDefinitionPointerFID);

	debugf("ServiceDefinition_toJava: Populating 'ServiceDefinition' from 'DP_SVC_DEF'...\n");
	jPointer = (*env)->GetLongField(env, jObject, jServiceDefinitionPointerFID);
	debugf("ServiceDefinition_toJava: The pointer is 0x%16.16X\n", jPointer);
	if (jPointer == 0L) {
		return; // there is nothing to do if we don't have a pointer...
	}
	pDpSvcDef = (DP_SVC_DEF *)jPointer;
	CHECK_SVC_DEF(pDpSvcDef)
	jCopyToDestination = pDpSvcDef->bCopyToDest;
	debugf("ServiceDefinition_toJava: Copy to destination is %s\n", (pDpSvcDef->bCopyToDest ? "true" : "false"));
	(*env)->SetBooleanField(env, jObject, jServiceDefinitionCopyToDestinationFID, jCopyToDestination);

	/* Handle the initialization operands.  We are going to run down the C linked list, create a new InitializationOperand
	 * object, populate its pointer, and add it to the ServiceDefinition's InitializationOperand list.
	 */
	jInitializationOperands = (*env)->GetObjectField(env, jObject, jServiceDefinitionInitializationOperandsFID);
	jInitializationOperandsSize = (*env)->CallIntMethod(env, jInitializationOperands, jArrayListSizeMID);
	/* If the list isn't empty, what do we do? I think we just add...
	 * The ServiceDefinition_toC will take care of any relinking and re-populating of the pDpSvcDef fields that might
	 * need to be done.  I'm not sure what duplicate parameters do to ODPP.  BUT, we can certainly nuke duplicates in
	 * Java land after calling ServiceDefinition_toJava().
	 */
	debugf("ServiceDefinition_toJava: Initialization operands size is %d.\n", jInitializationOperandsSize);
	pDpInitOpDef = pDpSvcDef->pParams;
	while(pDpInitOpDef != (DP_INIT_OP_DEF*) NULL) {
		jInitializationOperand = (*env)->NewObject(env, jInitializationOperandClass, jInitializationOperandConstructorMID);
		if (jInitializationOperand == NULL) {
			debugf("ServiceDefinition_toJava: InitializationOperand constructor failed\n");
			throwODPPException(env, "InitializationOperand constructor failed", ODPPFAILURE);
			return;
		}
		/* Populate the pointer.  Rely on InitializationOperand_toJava() to do anything else */
		jInitializationOperandPointer = (jlong) pDpInitOpDef;
		(*env)->SetLongField(env, jInitializationOperand, jInitializationOperandPointerFID, jInitializationOperandPointer);
		/* Add new object to list */
		debugf("ServiceDefinition_toJava: Adding InitializationOperand to list\n");
		addSucceeded = (*env)->CallBooleanMethod(env, jInitializationOperands, jArrayListAddMID, jInitializationOperand);
		if (!addSucceeded) {
			debugf("ServiceDefinition_toJava: Adding InitializationOperand to list failed\n");
			throwODPPException(env, "ServiceDefinition - add to InitializationOperand list failed", ODPPFAILURE);
			return;
		}
		/* Delete reference, advance to next item */
		(*env)->DeleteLocalRef(env, jInitializationOperand);
		pDpInitOpDef = pDpInitOpDef->pNext;
	}

	/* Same thing with the FieldDefinitions, but... we probably passed fields over, so if we have a populated list,
	 * don't add to it...
	 */

	jFieldDefinitions = (*env)->GetObjectField(env, jObject, jServiceDefinitionFieldDefinitionsFID);
	jFieldDefinitionsSize = (*env)->CallIntMethod(env, jFieldDefinitions, jArrayListSizeMID);
	debugf("ServiceDefinition_toJava: Field definitions size is %d.\n", jFieldDefinitionsSize);
	if (jFieldDefinitionsSize == 0) { // We have nothing, so populate...
		pDpFieldDef = pDpSvcDef->pFldDef;
		while (pDpFieldDef != (DP_FIELD_DEF *) NULL) {
			jFieldDefinition = (*env)->NewObject(env, jFieldDefinitionClass, jFieldDefinitionConstructorMID);
			if (jFieldDefinition == NULL) {
				debugf("ServiceDefinition_toJava: FieldDefinition constructor failed\n");
				throwODPPException(env, "FieldDefinition constructor failed", ODPPFAILURE);
				return;
			}
			/* Populate the pointer.  Rely on FieldDefinition_toJava() to do anything else */
			jFieldDefinitionPointer = (jlong) pDpFieldDef;
			(*env)->SetLongField(env, jFieldDefinition, jFieldDefinitionPointerFID, jFieldDefinitionPointer);
			/* Add new object to list */
			debugf("ServiceDefinition_toJava: Adding FieldDefinition to list\n");
			addSucceeded = (*env)->CallBooleanMethod(env, jFieldDefinitions, jArrayListAddMID, jFieldDefinition);
			if (!addSucceeded) {
				debugf("ServiceDefinition_toJava: Adding FieldDefinition to list failed\n");
				throwODPPException(env, "ServiceDefinition - add to FieldDefinition list failed", ODPPFAILURE);
				return;
			}
			/* Delete reference, advance to next item */
			(*env)->DeleteLocalRef(env, jFieldDefinition);
			pDpFieldDef = pDpFieldDef->pNext;
		}
	}
	debugf("ServiceDefinition_toJava:Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_internal_ServiceDefinition
 * Method:    free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_ServiceDefinition_free(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_SVC_DEF *pDpSvcDef;
    RETURN_IF_NULL(jServiceDefinitionPointerFID);
    debugf("ServiceDefinition_free: Freeing 'ServiceDefinition'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jServiceDefinitionPointerFID);
    debugf("ServiceDefinition_free: The pointer is 0x%16.16X\n", jPointer);
    /* We manually free everything, so we don't need or want to use ParmGen_Clean()     */
    if (jPointer != 0L) {
        pDpSvcDef = (DP_SVC_DEF *)jPointer;
        CHECK_SVC_DEF(pDpSvcDef)
        debugf("ServiceDefinition_free: Freeing 'DP_SVC_DEF' 0x%16.16X\n", jPointer);
        free(pDpSvcDef);
        jPointer = 0L;
        (*env)->SetLongField(env, jObject, jServiceDefinitionPointerFID, jPointer);
        debugf("ServiceDefinition_free: Free of 'ServiceDefinition' done.\n");
    }
    debugf("ServiceDefinition_free: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_internal_InitializationOperand
 * Method:    toC
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_InitializationOperand_toC(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_INIT_OP_DEF *pDpInitOpDef;
    jint jParameterId;
    jstring jParameterName;
    jsize jParameterNameLength;
    const jchar *jParameterNameChars;
    int jParameterNameSize;
    jstring jSValue;
    jsize jSValueLength;
    const jchar *jSValueChars;
    int jSValueSize;
    jint jIValue;
    RETURN_IF_NULL(jInitializationOperandPointerFID);
    debugf("InitializationOperand_toC: Populating 'DP_INIT_OP_DEF' structure from 'InitializationOperand'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jInitializationOperandPointerFID);
    debugf("InitializationOperand_toC: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        debugf("InitializationOperand_toC: Allocating 'InitializationOperand'...\n");
        pDpInitOpDef = (DP_INIT_OP_DEF *)malloc(sizeof(DP_INIT_OP_DEF));
        memset(pDpInitOpDef, 0, sizeof(DP_INIT_OP_DEF));
        INITIALIZE_ODPP_STRUCT_PTR(pDpInitOpDef, DP_INIT_OP_DEF)
        jPointer = (jlong) pDpInitOpDef;
        (*env)->SetLongField(env, jObject, jInitializationOperandPointerFID, jPointer);
        debugf("InitializationOperand_toC: Allocation of 'InitializationOperand' done. Pointer is 0x%16.16X\n", jPointer);
    }
    pDpInitOpDef = (DP_INIT_OP_DEF *)jPointer;
    CHECK_INIT_OP_DEF(pDpInitOpDef)
    jParameterId = (*env)->GetIntField(env, jObject, jInitializationOperandParameterIdFID);
    debugf("InitializationOperand_toC: The parameter id is %d\n", jParameterId);
    pDpInitOpDef->usParameterID = (unsigned short)jParameterId;
    jParameterName = (*env)->GetObjectField(env, jObject, jInitializationOperandParameterNameFID);
    if (jParameterName != NULL) {
        debugf("InitializationOperand_toC: Parameter name specified...\n");
        jParameterNameLength = (*env)->GetStringLength(env, jParameterName);
        debugf("InitializationOperand_toC: Parameter name is %d characters...\n", jParameterNameLength);
        jParameterNameChars = (*env)->GetStringChars(env, jParameterName, NULL);
        pDpInitOpDef->cParamNameSubType = 'W';
        pDpInitOpDef->iParamNameBufBytes = (jParameterNameLength + 1) * sizeof(ODPP_WCHAR);
        debugf("InitializationOperand_toC: Allocating 'DPPRM_NAME_WC_SS'...\n");
        pDpInitOpDef->PN.pWC = (DPPRM_NAME_WC_SS *)malloc(sizeof(DPPRM_NAME_WC_SS));
        debugf("InitializationOperand_toC: Allocated 'DPPRM_NAME_WC_SS'...0x%16.16X\n", (unsigned long long)pDpInitOpDef->PN.pWC);
        debugf("InitializationOperand_toC: Allocating 'ODPP_WCHAR' for parameter name...\n");
        pDpInitOpDef->PN.pWC->pParmName = malloc(pDpInitOpDef->iParamNameBufBytes);
        debugf("InitializationOperand_toC: Allocating 'ODPP_WCHAR' for parameter name...0x%16.16X\n", (unsigned long long)pDpInitOpDef->PN.pWC->pParmName);
        memset(pDpInitOpDef->PN.pWC->pParmName, 0, pDpInitOpDef->iParamNameBufBytes);
        // If wchar is 32-bit convert from Java UTF-16 to platform UTF-32, else use UTF-16 as is.
        if (cvtWchars) {
            Convert16To32(jParameterNameChars, jParameterNameLength, (ODPP_UINT *)pDpInitOpDef->PN.pWC->pParmName, jParameterNameLength, &jParameterNameSize);
        } else {
            WCSNCPY((ODPP_WCHAR *)pDpInitOpDef->PN.pWC->pParmName,(ODPP_WCHAR *)jParameterNameChars,jParameterNameLength);
        }
        debugf("InitializationOperand_toC: Parameter name is '%ls'\n", pDpInitOpDef->PN.pWC->pParmName);
        (*env)->ReleaseStringChars(env, jParameterName, jParameterNameChars);
        debugf("InitializationOperand_toC: Parameter name handled...\n");
    }
    jSValue = (*env)->GetObjectField(env, jObject, jInitializationOperandSValueIdFID);
    if (jSValue != NULL) {
        debugf("InitializationOperand_toC: The string value is not null, so handling string...\n");
        jSValueLength = (*env)->GetStringLength(env, jSValue);
        debugf("InitializationOperand_toC: String value is %d characters...\n", jSValueLength);
        jSValueChars = (*env)->GetStringChars(env, jSValue, NULL);
        pDpInitOpDef->iValueSubType = PARAM_VAL_WC;
        pDpInitOpDef->iValueBufBytes = (jSValueLength + 1) * sizeof(ODPP_WCHAR);
        debugf("InitializationOperand_toC: Allocating 'DPPRM_VAL_WC_SS'...\n");
        pDpInitOpDef->PV.pWC = (DPPRM_VAL_WC_SS *)malloc(sizeof(DPPRM_VAL_WC_SS));
        debugf("InitializationOperand_toC: Allocated 'DPPRM_VAL_WC_SS'...0x%16.16X\n", (unsigned long long)pDpInitOpDef->PV.pWC);
        debugf("InitializationOperand_toC: Allocating 'ODPP_WCHAR' for string value...\n");
        pDpInitOpDef->PV.pWC->pParamVal = malloc(pDpInitOpDef->iValueBufBytes);
        debugf("InitializationOperand_toC: Allocatied 'ODPP_WCHAR' for string value...0x%16.16X\n", (unsigned long long)pDpInitOpDef->PV.pWC->pParamVal);
        memset(pDpInitOpDef->PV.pWC->pParamVal, 0, pDpInitOpDef->iValueBufBytes);
        // If wchar is 32-bit convert from Java UTF-16 to platform UTF-32, else use UTF-16 as is.
        if (cvtWchars) {
            Convert16To32(jSValueChars, jSValueLength, (ODPP_UINT *)pDpInitOpDef->PV.pWC->pParamVal, jSValueLength, &jSValueSize);
        } else {
            WCSNCPY((ODPP_WCHAR *)pDpInitOpDef->PV.pWC->pParamVal, (ODPP_WCHAR *)jSValueChars, jSValueLength);
        }
        debugf("InitializationOperand_toC: String value is '%ls'\n", pDpInitOpDef->PV.pWC->pParamVal);
        (*env)->ReleaseStringChars(env, jSValue, jSValueChars);
        debugf("InitializationOperand_toC: String value handled...\n");
    } else {
        debugf("InitializationOperand_toC: The string value is null, so handling iValue...\n");
        jIValue = (*env)->GetIntField(env, jObject, jInitializationOperandIValueIdFID);
        debugf("InitializationOperand_toC: Integer value is %d\n", jIValue);
        pDpInitOpDef->iValueSubType = PARAM_VAL_NUM;
        pDpInitOpDef->PV.uiVal = (ODPP_UINT)jIValue;
    }
    debugf("InitializationOperand_toC: Done\n");

}

/*
 * Class:     com_ibm_nex_odpp_internal_InitializationOperand
 * Method:    toJava
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_InitializationOperand_toJava(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_INIT_OP_DEF *pDpInitOpDef;
    jint jParameterId;
    jstring jSource;
    jchar *jParameterNameChars;
    int jParameterNameSize;
    jchar *jSValueChars;
    int jSValueSize;
    jint jIValue;
    int maxNumChars;
    jchar tmpBuf[256];


    RETURN_IF_NULL(jInitializationOperandPointerFID);
    debugf("InitializationOperand_toJava: Populating InitializationOperand from 'DP_INIT_OP_DEF' structure...\n");
    jPointer = (*env)->GetLongField(env, jObject, jInitializationOperandPointerFID);
    debugf("InitializationOperand_toJava: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        // Nothing to do.
    	return;
    }
    pDpInitOpDef = (DP_INIT_OP_DEF *)jPointer;
    CHECK_INIT_OP_DEF(pDpInitOpDef)
    /* Set the Parameter ID */
    jParameterId = pDpInitOpDef->usParameterID;
    debugf("InitializationOperand_toJava: The parameter id is %d\n", jParameterId);
    (*env)->SetIntField(env, jObject, jInitializationOperandParameterIdFID, jParameterId);

    /* ParameterName is a bit tricky. We need to figure out if we have one, then we need to see what the encoding is
     * YES, I know the pointer test is a bit redundant.  It looks funny to me if I don't do it that way, so...
     */
    if (pDpInitOpDef->iParamNameBufBytes >0 && (pDpInitOpDef->PN.pWC != NULL || pDpInitOpDef->PN.pMC != NULL)) {
        debugf("InitializationOperand_toJava: We have a parameter Name\n");
    	/* We have a name.  Is it Wide, or Narrow? */
    	if (pDpInitOpDef->cParamNameSubType == 'W') {
            debugf("InitializationOperand_toJava: We have a WIDE parameter Name, \"%S\"\n", pDpInitOpDef->PN.pWC->pParmName);
    		maxNumChars = (pDpInitOpDef->iParamNameBufBytes/sizeof(ODPP_WCHAR)) - 1;
    		if ((maxNumChars+1)*sizeof(jchar) <= sizeof(tmpBuf)) {
    			jParameterNameChars = tmpBuf;
    		} else {
    			jParameterNameChars = (jchar *)malloc((maxNumChars + 1) * sizeof(jchar));
    		}
    		memset(jParameterNameChars, 0, (maxNumChars + 1) * sizeof(jchar));
    		/* WIDE */
            if (cvtWchars) {
            	Convert32To16((ODPP_UINT *)pDpInitOpDef->PN.pWC->pParmName, maxNumChars, jParameterNameChars, maxNumChars, &jParameterNameSize);
            } else {
                WCSNCPY((ODPP_WCHAR *)jParameterNameChars, (ODPP_WCHAR *)pDpInitOpDef->PN.pWC->pParmName,maxNumChars);
                jParameterNameSize = WCSLEN((ODPP_WCHAR *)jParameterNameChars);
            }
    	} else {
    		/* NARROW */
            debugf("InitializationOperand_toJava: We have a NARROW parameter Name, \"%s\"\n", pDpInitOpDef->PN.pMC->pParmName);
    		maxNumChars = pDpInitOpDef->iParamNameBufBytes - 1;
    		if ((maxNumChars+1)*sizeof(jchar) <= sizeof(tmpBuf)) {
    			jParameterNameChars = tmpBuf;
    		} else {
    			jParameterNameChars = (jchar *)malloc((maxNumChars + 1) * sizeof(jchar));
    		}
    		memset(jParameterNameChars, 0, (maxNumChars + 1) * sizeof(jchar));
    		/* Not sure how this accounts for the code page that is present in the DPPRM_NAME_MC
    		 * Judging by the bug in the header file, I don't think we use SBCS/MBCS names here.
    		 */
          	Convert8To16(jParameterNameChars, maxNumChars, (unsigned char *)pDpInitOpDef->PN.pMC->pParmName, maxNumChars, &jParameterNameSize);
    	}
        debugf("InitializationOperand_toJava: Setting parameter name on String\n");
     	jSource = (*env)->NewString(env, jParameterNameChars, jParameterNameSize);
        (*env)->SetObjectField(env, jObject, jInitializationOperandParameterNameFID, jSource);
       	if (jParameterNameChars != NULL && jParameterNameChars != tmpBuf) {
       		free(jParameterNameChars);
       	}
       	jParameterNameChars = NULL;
        (*env)->DeleteLocalRef(env, jSource);
    }

    /* Determine the type of the value */
    switch (pDpInitOpDef->iValueSubType) {

    	case PARAM_VAL_NUM:
    		/* Not bad -- we have a number. Set it and leave */
            debugf("InitializationOperand_toJava: We have an iValue...\n");
            jIValue = pDpInitOpDef->PV.uiVal;
            debugf("InitializationOperand_toJava: Integer value is %d\n", jIValue);
            (*env)->SetIntField(env, jObject, jInitializationOperandIValueIdFID, jIValue);
    		break;

    	case PARAM_VAL_WC:
    		/* A wide string... Do the Dance! */
            debugf("InitializationOperand_toJava: We have a wide string...\n");
    		maxNumChars = (pDpInitOpDef->iValueBufBytes/sizeof(ODPP_WCHAR)) - 1;
    		if ((maxNumChars + 1) * sizeof(jchar) <= sizeof(tmpBuf)) {
    			jSValueChars = tmpBuf;
    		} else {
    			jSValueChars = (jchar *)malloc((maxNumChars + 1) * sizeof(jchar));
    		}
    		memset(jSValueChars, 0, (maxNumChars + 1) * sizeof(jchar));
    		/* WIDE */
            if (cvtWchars) {
            	Convert32To16((ODPP_UINT *)pDpInitOpDef->PV.pWC->pParamVal, maxNumChars, jSValueChars, maxNumChars, &jSValueSize);
            } else {
                WCSNCPY((ODPP_WCHAR *)jSValueChars, (ODPP_WCHAR *)pDpInitOpDef->PV.pWC->pParamVal,maxNumChars);
                jSValueSize = WCSLEN((ODPP_WCHAR *)jSValueChars);
            }
            debugf("InitializationOperand_toJava: Setting WIDE parameter Value on String\n");
          	jSource = (*env)->NewString(env, jSValueChars, jSValueSize);
            (*env)->SetObjectField(env, jObject, jInitializationOperandSValueIdFID, jSource);
           	if (jSValueChars != NULL && jSValueChars != tmpBuf) {
           		free(jSValueChars);
           	}
           	jSValueChars = NULL;
            (*env)->DeleteLocalRef(env, jSource);
    		break;

    	case PARAM_VAL_MC:
    		/* OK, who's the wise guy who ordered this? */
            debugf("InitializationOperand_toJava: We have a narrow string...\n");
    		maxNumChars = pDpInitOpDef->iValueBufBytes - 1;
    		if ((maxNumChars + 1) * sizeof(jchar) <= sizeof(tmpBuf)) {
    			jSValueChars = tmpBuf;
    		} else {
    			jSValueChars = (jchar *)malloc((maxNumChars + 1) * sizeof(jchar));
    		}
    		memset(jSValueChars, 0, (maxNumChars + 1) * sizeof(jchar));
    		/* NARROW -- This doesn't account for code page. */
           	Convert8To16(jSValueChars, maxNumChars, (unsigned char *)pDpInitOpDef->PV.pMC->pParamVal, maxNumChars, &jSValueSize);
            debugf("InitializationOperand_toJava: Setting NARROW parameter Value on String\n");
          	jSource = (*env)->NewString(env, jSValueChars, jSValueSize);
            (*env)->SetObjectField(env, jObject, jInitializationOperandSValueIdFID, jSource);
           	if (jSValueChars != NULL && jSValueChars != tmpBuf) {
           		free(jSValueChars);
           	}
           	jSValueChars = NULL;
            (*env)->DeleteLocalRef(env, jSource);
    		break;

    	case PARAM_VAL_NONE:
    		/* FALLTHROUGH */
    	default:
    		/* Nothing else to do. It's Miller Time! */
            debugf("InitializationOperand_toJava: We have a no value...\n");
    		break;
    }
    debugf("InitializationOperand_toJava: Done\n");
}

/*
 * Class:     com_ibm_nex_odpp_internal_InitializationOperand
 * Method:    free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_InitializationOperand_free(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_INIT_OP_DEF *pDpInitOpDef;
    RETURN_IF_NULL(jInitializationOperandPointerFID);
    debugf("InitializationOperand_free: Freeing 'InitializationOperand'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jInitializationOperandPointerFID);
    debugf("InitializationOperand_free: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer != 0L) {
        pDpInitOpDef = (DP_INIT_OP_DEF *)jPointer;
        CHECK_INIT_OP_DEF(pDpInitOpDef)
        if (pDpInitOpDef->PN.pWC != NULL) {
            if (pDpInitOpDef->PN.pWC->pParmName != NULL) {
                debugf("InitializationOperand_free: Freeing 'ODPP_WCHAR' for parameter name: 0x%16.16X\n", (unsigned long long)pDpInitOpDef->PN.pWC->pParmName);
                free(pDpInitOpDef->PN.pWC->pParmName);
                pDpInitOpDef->PN.pWC->pParmName = (ODPP_WCHAR *)NULL;
            }
            debugf("InitializationOperand_free: Freeing 'DPPRM_NAME_WC_SS'...0x%16.16X\n", (unsigned long long) pDpInitOpDef->PN.pWC);
            free(pDpInitOpDef->PN.pWC);
        }
        if (pDpInitOpDef->PV.pWC != NULL) {
            if (pDpInitOpDef->iValueSubType > 1) {
                if (pDpInitOpDef->PV.pWC->pParamVal != NULL) {
                    debugf("InitializationOperand_free: Freeing 'ODPP_WCHAR' for string value...0x%16.16X\n", (unsigned long long) pDpInitOpDef->PV.pWC->pParamVal);
                    free(pDpInitOpDef->PV.pWC->pParamVal);
                    pDpInitOpDef->PV.pWC->pParamVal = NULL;
                }
                debugf("InitializationOperand_free: Freeing 'DPPRM_VAL_WC_SS'...0x%16.16X\n", (unsigned long long)pDpInitOpDef->PV.pWC);
                free(pDpInitOpDef->PV.pWC);
                pDpInitOpDef->PV.pWC = NULL;
            } else {
                debugf("InitializationOperand_free: The value sub type is 1 or less, so cannot free 'DPPRM_VAL_WC_SS' as it's really the integer value\n");
            }
        }
        debugf("InitializationOperand_free: Freeing 'DP_INIT_OP_DEF'...0x%16.16X\n", (unsigned long long)pDpInitOpDef);
        free(pDpInitOpDef);
        jPointer = 0L;
        (*env)->SetLongField(env, jObject, jInitializationOperandPointerFID, jPointer);
        debugf("InitializationOperand_free: Free of 'InitializationOperand' done.\n");
    }
    debugf("InitializationOperand_free: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_internal_FieldDefinition
 * Method:    toC
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_FieldDefinition_toC(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_FIELD_DEF *pDpFieldDef;
    jint jId;
    jint jLength;
    jint jPrecision;
    jint jScale;
    jstring jColumnName;
    jsize jColumnNameLength;
    const jchar *jColumnNameChars;
    int jColumnNameSize;
    RETURN_IF_NULL(jFieldDefinitionPointerFID);
    debugf("FieldDefinition_toC: Populating 'DP_FIELD_DEF' from 'FieldDefinition'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jFieldDefinitionPointerFID);
    debugf("FieldDefinition_toC: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        debugf("FieldDefinition_toC: Allocating 'FieldDefinition'...\n");
        debugf("FieldDefinition_toC: Allocating 'DP_FIELD_DEF'...\n");
        pDpFieldDef = (DP_FIELD_DEF *)malloc(sizeof(DP_FIELD_DEF));
        memset(pDpFieldDef, 0, sizeof(DP_FIELD_DEF));
        INITIALIZE_ODPP_STRUCT_PTR(pDpFieldDef, DP_FIELD_DEF)
        debugf("FieldDefinition_toC: Allocating 'DPFD_WC_SS'...\n");
        pDpFieldDef->CN.pWC = (DPFD_WC_SS *)malloc(sizeof(DPFD_WC_SS));
        debugf("FieldDefinition_toC: Allocated 'DPFD_WC_SS'...0x%16.16X\n", (unsigned long long)pDpFieldDef->CN.pWC);
        memset(pDpFieldDef->CN.pWC, 0, sizeof(DPFD_WC_SS));
        jPointer = (jlong) pDpFieldDef;
        (*env)->SetLongField(env, jObject, jFieldDefinitionPointerFID, jPointer);
        debugf("FieldDefinition_toC: Allocation of 'FieldDefinition' done. Pointer is 0x%16.16X\n", jPointer);
    }
    pDpFieldDef = (DP_FIELD_DEF *)jPointer;
    CHECK_FIELD_DEF(pDpFieldDef)
    jId = (*env)->GetIntField(env, jObject, jFieldDefinitionIdFID);
    pDpFieldDef->sDatatype = (short)jId;
    debugf("FieldDefinition_toC: Data type is %d.\n", pDpFieldDef->sDatatype);
    jLength = (*env)->GetIntField(env, jObject, jFieldDefinitionLengthFID);
    pDpFieldDef->iLength = jLength;
    debugf("FieldDefinition_toC: Length is %d.\n", pDpFieldDef->iLength);
    jPrecision = (*env)->GetIntField(env, jObject, jFieldDefinitionPrecisionFID);
    pDpFieldDef->iPrecision = jPrecision;
    debugf("FieldDefinition_toC: Precision is %d.\n", pDpFieldDef->iPrecision);
    jScale = (*env)->GetIntField(env, jObject, jFieldDefinitionScaleFID);
    pDpFieldDef->sScale = (short)jScale;
    debugf("FieldDefinition_toC: Scale is %d.\n", pDpFieldDef->sScale);
    jColumnName = (*env)->GetObjectField(env, jObject, jFieldDefinitionColumnNameFID);
    jColumnNameLength = (*env)->GetStringLength(env, jColumnName);
    pDpFieldDef->iColNameBytes = (jColumnNameLength + 1) * sizeof(ODPP_WCHAR);
    if (pDpFieldDef->CN.pWC->pColName != NULL) {
        debugf("FieldDefinition_toC: Freeing column name...0x%16.16X\n", (unsigned long long) pDpFieldDef->CN.pWC->pColName);
        free(pDpFieldDef->CN.pWC->pColName);
        pDpFieldDef->CN.pWC->pColName = NULL;
    }
    debugf("FieldDefinition_toC: Allocating column name...\n");
    pDpFieldDef->CN.pWC->pColName = malloc(pDpFieldDef->iColNameBytes);
    debugf("FieldDefinition_toC: Allocated column name...0x%16.16X\n", (unsigned long long)pDpFieldDef->CN.pWC->pColName);
    memset(pDpFieldDef->CN.pWC->pColName, 0, pDpFieldDef->iColNameBytes);
    jColumnNameChars = (*env)->GetStringChars(env, jColumnName, NULL);
    // If wchar is 32-bit convert from Java UTF-16 to platform UTF-32, else use UTF-16 as is.
    if (cvtWchars) {
        Convert16To32(jColumnNameChars, jColumnNameLength, (ODPP_UINT *)pDpFieldDef->CN.pWC->pColName, jColumnNameLength, &jColumnNameSize);
    } else {
        WCSNCPY((ODPP_WCHAR *)pDpFieldDef->CN.pWC->pColName, (ODPP_WCHAR *)jColumnNameChars, jColumnNameLength);
    }
    (*env)->ReleaseStringChars(env, jColumnName, jColumnNameChars);
    debugf("FieldDefinition_toC: Column name is '%ls'.\n", pDpFieldDef->CN.pWC->pColName);
    debugf("FieldDefinition_toC: Done.\n");

}

/*
 * Class:     com_ibm_nex_odpp_internal_FieldDefinition
 * Method:    toJava
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_FieldDefinition_toJava(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_FIELD_DEF *pDpFieldDef;
    jint jId;
    jint jLength;
    jint jPrecision;
    jint jScale;
    jchar *jColumnNameChars;
    int jColumnNameSize;
    int maxNumChars;
    jstring jSource;
    jchar tmpBuf[256];

    RETURN_IF_NULL(jFieldDefinitionPointerFID);

    debugf("FieldDefinition_toJava: Populating 'FieldDefinition' from 'DP_FIELD_DEF'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jFieldDefinitionPointerFID);
    debugf("FieldDefinition_toJava: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        // If pointer is 0, there is nothing to do here.  Bail.
    	return;
    }
    pDpFieldDef = (DP_FIELD_DEF *)jPointer;
    CHECK_FIELD_DEF(pDpFieldDef)
    /* Set ID */
    jId = pDpFieldDef->sDatatype;
    debugf("FieldDefinition_toJava: Setting datatype ID %d\n", pDpFieldDef->sDatatype);
    (*env)->SetIntField(env, jObject, jFieldDefinitionIdFID, jId);

    /* Set Length */
    jLength = pDpFieldDef->iLength;
    debugf("FieldDefinition_toJava: Setting Length to %d.\n", pDpFieldDef->iLength);
    (*env)->SetIntField(env, jObject, jFieldDefinitionLengthFID, jLength);

    /* Set Precision */
    jPrecision = pDpFieldDef->iPrecision;
    debugf("FieldDefinition_toJava: Setting Precision to %d.\n", pDpFieldDef->iPrecision);
    (*env)->SetIntField(env, jObject, jFieldDefinitionPrecisionFID, jPrecision);

    /* Set Scale */
    jScale = pDpFieldDef->sScale;
    debugf("FieldDefinition_toJava: Setting Scale to %d.\n", pDpFieldDef->sScale);
    (*env)->SetIntField(env, jObject, jFieldDefinitionScaleFID, jScale);

    /* Now handle column name */
    debugf("FieldDefinition_toJava: Working on Column Name.\n");
    if (pDpFieldDef->iColNameBytes > 0 && (pDpFieldDef->CN.pMC->pColName != NULL || pDpFieldDef->CN.pWC->pColName != NULL) ) {
		if (pDpFieldDef->cSubType == 'M') {
			/* NARROW */
			maxNumChars = pDpFieldDef->iColNameBytes - 1;
		    debugf("FieldDefinition_toJava: NARROW maxNumChars is %d.\n", maxNumChars);
		    if ((maxNumChars + 1) * sizeof(jchar) <= sizeof(tmpBuf)){
		    	jColumnNameChars = tmpBuf;
		    } else {
		    	jColumnNameChars = (jchar *)malloc(sizeof(jchar) * (maxNumChars + 1));
		    }
			memset(jColumnNameChars, 0, (maxNumChars + 1) * sizeof(jchar));
			/* NARROW -- This doesn't account for code page. */
			Convert8To16(jColumnNameChars, maxNumChars, (unsigned char *)pDpFieldDef->CN.pMC->pColName, maxNumChars, &jColumnNameSize);
			debugf("FieldDefinition_toJava: Setting NARROW parameter Value on String\n");
		} else {
			/* WIDE */
			maxNumChars = (pDpFieldDef->iColNameBytes/sizeof(ODPP_WCHAR)) - 1;
		    debugf("FieldDefinition_toJava: WIDE maxNumChars is %d.\n", maxNumChars);
		    if ((maxNumChars + 1) * sizeof(jchar) <= sizeof(tmpBuf)){
		    	jColumnNameChars = tmpBuf;
		    } else {
		    	jColumnNameChars = (jchar *)malloc(sizeof(jchar) * (maxNumChars + 1));
		    }
			jColumnNameChars = (jchar *)malloc((maxNumChars + 1) * sizeof(jchar));
			memset(jColumnNameChars, 0, (maxNumChars + 1) * sizeof(jchar));
			if (cvtWchars) {
				Convert32To16((ODPP_UINT *)pDpFieldDef->CN.pWC->pColName, maxNumChars, jColumnNameChars, maxNumChars, &jColumnNameSize);
			} else {
				WCSNCPY((ODPP_WCHAR *)jColumnNameChars, (ODPP_WCHAR *)pDpFieldDef->CN.pWC->pColName,maxNumChars);
				jColumnNameSize = WCSLEN((ODPP_WCHAR *)jColumnNameChars);
			}
			debugf("FieldDefinition_toJava: Setting WIDE parameter Value on String\n");
		}
		jSource = (*env)->NewString(env, jColumnNameChars, jColumnNameSize);
		(*env)->SetObjectField(env, jObject, jFieldDefinitionColumnNameFID, jSource);
		if (jColumnNameChars != NULL && jColumnNameChars != tmpBuf) {
			free(jColumnNameChars);
		}
		jColumnNameChars = NULL;
		(*env)->DeleteLocalRef(env, jSource);
	}
    debugf("FieldDefinition_toJava: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_internal_FieldDefinition
 * Method:    free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_internal_FieldDefinition_free(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_FIELD_DEF *pDpFieldDef;
    RETURN_IF_NULL(jFieldDefinitionPointerFID);
    debugf("FieldDefinition_free: Freeing 'FieldDefinition'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jFieldDefinitionPointerFID);
    debugf("FieldDefinition_free: The pointer is0x%16.16X\n", jPointer);
    if (jPointer != 0L) {
        pDpFieldDef = (DP_FIELD_DEF *)jPointer;
        CHECK_FIELD_DEF(pDpFieldDef)
        if (pDpFieldDef->CN.pWC != NULL) {
            if (pDpFieldDef->CN.pWC->pColName != NULL) {
                debugf("FieldDefinition_free: Freeing column name...0x%16.16X\n", (unsigned long long)pDpFieldDef->CN.pWC->pColName);
                free(pDpFieldDef->CN.pWC->pColName);
                pDpFieldDef->CN.pWC->pColName = NULL;
            }
            debugf("FieldDefinition_free: Freeing 'DPFD_WC_SS'...0x%16.16X\n", (unsigned long long)pDpFieldDef->CN.pWC);
            free(pDpFieldDef->CN.pWC);
            pDpFieldDef->CN.pWC = NULL;
        }
        debugf("FieldDefinition_free: Freeing 'DP_FIELD_DEF'...\n");
        free(pDpFieldDef);
        jPointer = 0L;
        (*env)->SetLongField(env, jObject, jFieldDefinitionPointerFID, jPointer);
        debugf("FieldDefinition_free: Free of 'FieldDefinition' done.\n");
    }
    debugf("FieldDefinition_free: Done.\n");

}

/*
 * Class:     com_ibm_nex_odpp_RowSet
 * Method:    toC
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_RowSet_toC(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_ROWSET_DEF *pDpRowSetDef;
    jobject jRows;
    jsize jRowCount;
    DP_ROW_DEF *pCurrentDpRowDef;
    int i;
    jobject jRow;
    jlong jRowPointer;
    DP_ROW_DEF *pDpRowDef;
    RETURN_IF_NULL(jRowSetPointerFID);
    debugf("RowSet_toC: Populating 'DP_ROWSET_DEF' from 'Rowset'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jRowSetPointerFID);
    debugf("RowSet_toC: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        debugf("RowSet_toC: Allocating 'Rowset'...\n");
        debugf("RowSet_toC: Allocating 'DP_ROWSET_DEF'...\n");
        pDpRowSetDef = (DP_ROWSET_DEF *)malloc(sizeof(DP_ROWSET_DEF));
        debugf("RowSet_toC: Allocated 'DP_ROWSET_DEF'...0x%16.16X\n", (unsigned long long) pDpRowSetDef);
        memset(pDpRowSetDef, 0, sizeof(DP_ROWSET_DEF));
        INITIALIZE_ODPP_STRUCT_PTR(pDpRowSetDef, DP_ROWSET_DEF)
        jPointer = (jlong) pDpRowSetDef;
        (*env)->SetLongField(env, jObject, jRowSetPointerFID, jPointer);
        debugf("RowSet_toC: Allocation of 'Rowset' done. Pointer is 0x%16.16X\n", jPointer);
    }
    pDpRowSetDef = (DP_ROWSET_DEF *)jPointer;
    CHECK_ROWSET_DEF(pDpRowSetDef)
    jRows = (*env)->GetObjectField(env, jObject, jRowSetRowsFID);
    jRowCount = (*env)->GetArrayLength(env, jRows);
    debugf("RowSet_toC: There are %d rows.\n", jRowCount);
    pDpRowSetDef->iCount = jRowCount;
    if (jRowCount > 0) {
        debugf("RowSet_toC: Linking rows...\n");
        pCurrentDpRowDef = NULL;
        for (i = 0; i < jRowCount; i++) {
            jRow = (*env)->GetObjectArrayElement(env, jRows, i);
            jRowPointer = (*env)->GetLongField(env, jRow, jRowPointerFID);
            pDpRowDef = (DP_ROW_DEF *)jRowPointer;
            CHECK_ROW_DEF(pDpRowDef)
            if (pCurrentDpRowDef == NULL) {
                pDpRowSetDef->pRowDefine = pDpRowDef;
            } else {
                pCurrentDpRowDef->pNext = pDpRowDef;
            }
            pCurrentDpRowDef = pDpRowDef;
            (*env)->DeleteLocalRef(env, jRow);
        }
        pCurrentDpRowDef->pNext = (DP_ROW_DEF *)NULL;
    }
    (*env)->DeleteLocalRef(env, jRows);
    /* No need to do anything with rowSetHasInfo here.  It is an OUT parameter */
    debugf("RowSet_toC: Done.\n");

}

/*
 * Class:     com_ibm_nex_odpp_RowSet
 * Method:    toJava
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_RowSet_toJava(JNIEnv *env, jobject jObject)
{
	// Do nothing here.
}

/*
 * Class:     com_ibm_nex_odpp_RowSet
 * Method:    free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_RowSet_free(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_ROWSET_DEF *pDpRowsetDef;
    RETURN_IF_NULL(jRowSetPointerFID);
    debugf("RowSet_free : Freeing 'Rowset'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jRowSetPointerFID);
    debugf("RowSet_free: Freeing 'DP_ROWSET_DEF'.0x%16.16X\n", jPointer);
    if (jPointer != 0L) {
        pDpRowsetDef = (DP_ROWSET_DEF *)jPointer;
        CHECK_ROWSET_DEF(pDpRowsetDef)
        free(pDpRowsetDef);
        jPointer = 0L;
        (*env)->SetLongField(env, jObject, jRowSetPointerFID, jPointer);
        debugf("RowSet_free: Free of 'Rowset' done.\n");
    }
    debugf("RowSet_free: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_Row
 * Method:    toC
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_Row_toC(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_ROW_DEF *pDpRowDef;
	jboolean jHasError;
    jobject jFields;
    jsize jFieldCount;
    DP_FIELD_DATA_DEF *pCurrentDpFieldDataDef;
    int i;
    jobject jField;
    jlong jFieldPointer;
    DP_FIELD_DATA_DEF *pDpFieldDataDef;
    RETURN_IF_NULL(jRowPointerFID);
    debugf("Row_toC: Populating 'DP_ROW_DEF' from 'Row'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jRowPointerFID);
    debugf("Row_toC: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        debugf("Row_toC: Allocating 'Row'...\n");
        debugf("Row_toC: Allocating 'DP_ROW_DEF'...\n");
        pDpRowDef = (DP_ROW_DEF *)malloc(sizeof(DP_ROW_DEF));
        memset(pDpRowDef, 0, sizeof(DP_ROW_DEF));
        INITIALIZE_ODPP_STRUCT_PTR(pDpRowDef, DP_ROW_DEF)
        jPointer = (jlong) pDpRowDef;
        (*env)->SetLongField(env, jObject, jRowPointerFID, jPointer);
        debugf("Row_toC: Allocation of 'Row' done. Pointer is 0x%16.16X\n", jPointer);
    }
    pDpRowDef = (DP_ROW_DEF *)jPointer;
    CHECK_ROW_DEF(pDpRowDef)
    jHasError = (*env)->GetBooleanField(env, jObject, jRowHasErrorFID);
    pDpRowDef->bHasError = jHasError;
    jFields = (*env)->GetObjectField(env, jObject, jRowFieldsFID);
    jFieldCount = (*env)->GetArrayLength(env, jFields);
    debugf("Row_toC: There are %d fields.\n", jFieldCount);
    pDpRowDef->sCount = (short)jFieldCount;
    if (jFieldCount > 0) {
        debugf("Row_toC: Linking fields...\n");
        pCurrentDpFieldDataDef = NULL;
        for (i = 0; i < jFieldCount; i++) {
            jField = (*env)->GetObjectArrayElement(env, jFields, i);
            jFieldPointer = (*env)->GetLongField(env, jField, jFieldPointerFID);
            pDpFieldDataDef = (DP_FIELD_DATA_DEF *)jFieldPointer;
            CHECK_FIELD_DATA_DEF(pDpFieldDataDef)
            if (pCurrentDpFieldDataDef == NULL) {
                pDpRowDef->pFldDataDefine = pDpFieldDataDef;
            } else {
                pCurrentDpFieldDataDef->pNext = pDpFieldDataDef;
            }
            pCurrentDpFieldDataDef = pDpFieldDataDef;
            (*env)->DeleteLocalRef(env, jField);
        }
        pCurrentDpFieldDataDef->pNext = (DP_FIELD_DATA_DEF *)NULL;
    }
    (*env)->DeleteLocalRef(env, jFields);
    debugf("Row_toC: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_Row
 * Method:    toJava
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_Row_toJava(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_ROW_DEF *pDpRowDef;
    jboolean jHasError;
    RETURN_IF_NULL(jRowPointerFID);
    debugf("Row_toJava: Populating 'Row' from 'DP_ROW_DEF'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jRowPointerFID);
    debugf("Row_toJava: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer != 0L) {
        pDpRowDef = (DP_ROW_DEF *)jPointer;
        CHECK_ROW_DEF(pDpRowDef)
        jHasError = pDpRowDef->bHasError;
        (*env)->SetBooleanField(env, jObject, jRowHasErrorFID, jHasError);
    }
    debugf("Row_toJava: Done\n");

}

/*
 * Class:     com_ibm_nex_odpp_Row
 * Method:    free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_Row_free(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_ROW_DEF *pDpRowDef;
    RETURN_IF_NULL(jRowPointerFID);
    debugf("Row_free: Freeing 'Row'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jRowPointerFID);
    debugf("Row_free: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer != 0L) {
        pDpRowDef = (DP_ROW_DEF *)jPointer;
        CHECK_ROW_DEF(pDpRowDef)
        debugf("Row_free: Freeing 'DP_ROW_DEF'...0x%16.16X\n",(unsigned long long) pDpRowDef);
        free(pDpRowDef);
        jPointer = 0L;
        (*env)->SetLongField(env, jObject, jRowPointerFID, jPointer);
        debugf("Row_free: Free of 'Row' done.\n");
    }
    debugf("Row_free: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_Field
 * Method:    toC
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_Field_toC(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_FIELD_DATA_DEF *pDpFieldDataDef;
    jint jId;
    jint jSourceSize;
    jobject jSource;
    jstring jString;
    jsize jStringLength;
    const jchar *jChars;
    int jStringSize;
    jint jErrorCode;
    RETURN_IF_NULL(jFieldPointerFID);
    debugf("Field_toC: Populating 'DP_FIELD_DATA_DEF' from 'Field'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jFieldPointerFID);
    debugf("Field_toC: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer == 0L) {
        debugf("Field_toC: Allocating 'Field'...\n");
        debugf("Field_toC: Allocating 'DP_FIELD_DATA_DEF'...\n");
        pDpFieldDataDef = (DP_FIELD_DATA_DEF *)malloc(sizeof(DP_FIELD_DATA_DEF));
        memset(pDpFieldDataDef, 0, sizeof(DP_FIELD_DATA_DEF));
        INITIALIZE_ODPP_STRUCT_PTR(pDpFieldDataDef, DP_FIELD_DATA_DEF)
        jPointer = (jlong) pDpFieldDataDef;
        (*env)->SetLongField(env, jObject, jFieldPointerFID, jPointer);
        debugf("Field_toC: Allocation of 'Field' done. Pointer is 0x%16.16X\n", jPointer);
    }
    pDpFieldDataDef = (DP_FIELD_DATA_DEF *)jPointer;
    CHECK_FIELD_DATA_DEF(pDpFieldDataDef)
    jId = (*env)->GetIntField(env, jObject, jFieldIdFID);
    debugf("Field_toC: The data type id is %d.\n", jId);
    jSourceSize = (*env)->GetIntField(env, jObject, jFieldSourceSizeFID);
    debugf("Field_toC: The source size is %d bytes.\n", jSourceSize);
    if ((pDpFieldDataDef->pSrcBuf == NULL) || (pDpFieldDataDef->iSrcBufLen < jSourceSize)) {
        if (pDpFieldDataDef->pSrcBuf != NULL) {
            debugf("Field_toC: Freeing pSrcBuf for reallocation. Pointer is 0x%16.16X\n", (unsigned long long) pDpFieldDataDef->pSrcBuf);
            free(pDpFieldDataDef->pSrcBuf);
            pDpFieldDataDef->pSrcBuf = NULL;
        }
        pDpFieldDataDef->iSrcBufLen = jSourceSize;
        if (pDpFieldDataDef->iSrcBufLen > 0) {
            pDpFieldDataDef->pSrcBuf = malloc(pDpFieldDataDef->iSrcBufLen);
            debugf("Field_toC: Allocation of pSrcBuf done %d bytes. Pointer is 0x%16.16X\n",jSourceSize, (unsigned long long) pDpFieldDataDef->pSrcBuf);
            //memset(pDpFieldDataDef->pSrcBuf, 0, pDpFieldDataDef->iSrcBufLen);
        } else {
            pDpFieldDataDef->pSrcBuf = NULL;
        }
    }
    if (pDpFieldDataDef->pSrcBuf != NULL) {
    	memset(pDpFieldDataDef->pSrcBuf, 0, pDpFieldDataDef->iSrcBufLen);
    }
    jSource = (*env)->GetObjectField(env, jObject, jFieldSourceFID);
    if (jSource != NULL) {
        pDpFieldDataDef->bSrcNull = FALSE;
        if ((*env)->IsInstanceOf(env, jSource, jStringClass)) {
            debugf("Field_toC: The source is a string...\n");
            jString = (jstring)jSource;
			jStringLength = (*env)->GetStringLength(env, jString);
            jChars = (*env)->GetStringChars(env, jString, NULL);
            switch (jId) {
                case ODPPDATATYPE_CHAR:
                case ODPPDATATYPE_VARCHAR:
                case ODPPDATATYPE_VARCHAR_SZ:
                    Convert16To8(pDpFieldDataDef->pSrcBuf, pDpFieldDataDef->iSrcBufLen, jChars, jStringLength, &jStringSize);
                    break;
                    
                case ODPPDATATYPE_WCHAR:
                case ODPPDATATYPE_WVARCHAR:
                case ODPPDATATYPE_WVARCHAR_SZ:
                    // If wchar is 32-bit convert from Java UTF-16 to platform UTF-32, else use UTF-16 as is.
                    if (cvtWchars) {
                        Convert16To32(jChars, jStringLength, (ODPP_UINT *)pDpFieldDataDef->pSrcBuf, (pDpFieldDataDef->iSrcBufLen / sizeof(ODPP_WCHAR))-1, &jStringSize);
                    } else {
						//WCSNCPY((ODPP_WCHAR *)pDpFieldDataDef->pSrcBuf, jChars, jStringLength);
                    	//(void)printf("pre-wcsncpy. jStringLength = %d, pDpFieldDataDef->iSrcBufLen is %d\n", jStringLength, pDpFieldDataDef->iSrcBufLen);
                    	//(void)printf("wcsncpy(0x%16.16X, 0x%16.16X, %d)\n", (unsigned long long)pDpFieldDataDef->pSrcBuf,(unsigned long long) jChars, (pDpFieldDataDef->iSrcBufLen/sizeof(ODPP_WCHAR))-1);
                    	WCSNCPY((ODPP_WCHAR *)pDpFieldDataDef->pSrcBuf, (ODPP_WCHAR *)jChars, min(jStringLength,(pDpFieldDataDef->iSrcBufLen/sizeof(ODPP_WCHAR))-1));
                   }
                    break;
            }
            (*env)->ReleaseStringChars(env, jString, jChars);
        }
        (*env)->DeleteLocalRef(env, jSource);
    } else {
        debugf("Field_toC: Source is NULL.\n");
        pDpFieldDataDef->bSrcNull = TRUE;
    }
#ifdef IOQJNI_USE_DESTINATION
    jDestinationSize = (*env)->GetIntField(env, jObject, jFieldDestinationSizeFID);
    debugf("Field_toC: The destination size is %d bytes.\n", jDestinationSize);
    if ((pDpFieldDataDef->pDstBuf == NULL) || (pDpFieldDataDef->iDstBufLen < jDestinationSize)) {
        if (pDpFieldDataDef->pDstBuf != NULL) {
            debugf("Field_toC: Freeing pDstBuf for reallocation. Pointer is 0x%16.16X\n", (unsigned long long) pDpFieldDataDef->pDstBuf);
            free(pDpFieldDataDef->pDstBuf);
            pDpFieldDataDef->pDstBuf = NULL;
        }
        pDpFieldDataDef->iDstBufLen = jDestinationSize;
        if (pDpFieldDataDef->iDstBufLen > 0) {
            pDpFieldDataDef->pDstBuf = malloc(pDpFieldDataDef->iDstBufLen);
            debugf("Field_toC: Allocation of pDstBuf done %d bytes. Pointer is 0x%16.16X\n",jDestinationSize, (unsigned long long) pDpFieldDataDef->pDstBuf);
            memset(pDpFieldDataDef->pDstBuf, 0, pDpFieldDataDef->iDstBufLen);
        } else {
            pDpFieldDataDef->pDstBuf = NULL;
        }
    }
# else
    pDpFieldDataDef->pDstBuf = NULL;
#endif
    jErrorCode = (*env)->GetIntField(env, jObject, jFieldErrorCodeFID);
    pDpFieldDataDef->iErrorCode = jErrorCode;
    // No need to propagate the error message here.  That is set on return.

    debugf("Field_toC: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_Field
 * Method:    toJava
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_Field_toJava(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_FIELD_DATA_DEF *pDpFieldDataDef;
    jint jId;
    jsize jStringLength;
    int jStringSize;
    jchar *jChars;
    jstring jSource;
    jint jErrorCode;
    jchar tmpBuf[256];
    RETURN_IF_NULL(jFieldPointerFID);
    debugf("Field_toJava: Populating 'Field' from 'DP_FIELD_DATA_DEF'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jFieldPointerFID);
    debugf("Field_toJava: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer != 0L) {
        pDpFieldDataDef = (DP_FIELD_DATA_DEF *)jPointer;
        CHECK_FIELD_DATA_DEF(pDpFieldDataDef)
        jId = (*env)->GetIntField(env, jObject, jFieldIdFID);
        jStringLength = 0;
        jStringSize = 0;
        jErrorCode = pDpFieldDataDef->iErrorCode;
        (*env)->SetIntField(env, jObject, jFieldErrorCodeFID, jErrorCode);
        debugf("Field_toJava: The data type id is %d.\n", jId);
        if (pDpFieldDataDef->bSrcNull) {
            debugf("Field_toJava: Source is null.\n");
            (*env)->SetObjectField(env, jObject, jFieldSourceFID, NULL);
        } else {
            if (pDpFieldDataDef->pSrcBuf != NULL) {
                switch (jId) {
                    case ODPPDATATYPE_CHAR:
                    case ODPPDATATYPE_VARCHAR:
                    case ODPPDATATYPE_VARCHAR_SZ:
                        debugf("Field_toJava: The source data type is a 'CHAR' variation.\n");
#ifdef IOQJNI_SUPPORT_CHAR_TYPES
                        jStringLength = pDpFieldDataDef->iSrcBufLen / sizeof(ODPP_CHAR);
                        debugf("Field_toJava: Source string length is %d.\n", jStringLength);
                        jChars = (jchar *)malloc(sizeof(jchar) * (pDpFieldDataDef->iSrcBufLen + 1));
                      	Convert8To16(jChars, jStringLength, (ODPP_CHAR *)pDpFieldDataDef->pSrcBuf, jStringLength, &jStringSize);
                      	jSource = (*env)->NewString(env, jChars, jStringLength);
                       	free(jChars);
                       	jChars = NULL;
                        (*env)->SetObjectField(env, jObject, jFieldSourceFID, jSource);
                        (*env)->DeleteLocalRef(env, jSource);
#endif
                        break;
                        
                    case ODPPDATATYPE_WCHAR:
                    case ODPPDATATYPE_WVARCHAR:
                    case ODPPDATATYPE_WVARCHAR_SZ:
                        debugf("Field_toJava: The source data type is a 'WCHAR' variation.\n");
                        if (jId == ODPPDATATYPE_WVARCHAR_SZ) {
                            jStringLength = pDpFieldDataDef->iSrcBufLen / sizeof(ODPP_WCHAR);
                            *((ODPP_WCHAR*)pDpFieldDataDef->pSrcBuf+(jStringLength - 1)) = L'\0';
                            jStringLength = (jsize) WCSLEN((ODPP_WCHAR *)pDpFieldDataDef->pSrcBuf);
                        } else {
                            jStringLength = pDpFieldDataDef->iSrcBufLen / sizeof(ODPP_WCHAR);
                        }
                        debugf("Field_toJava: Source string length is %d chars. SrcBufLen is %d bytes\n", jStringLength, pDpFieldDataDef->iSrcBufLen);
                        // If wchar is 32-bit convert from platform UTF-32 to Java UTF-16, else use UTF-16 as is.
                        if (cvtWchars) {
                        	if ((jStringLength + 1) * sizeof(jchar) <= sizeof(tmpBuf)){
                        		jChars = tmpBuf;
                        	} else {
                        		jChars = (jchar *)malloc(sizeof(jchar) * (jStringLength + 1));
                        	}
                        	memset(jChars,0,(jStringLength + 1) * sizeof(jchar));
                        	Convert32To16((ODPP_UINT *)pDpFieldDataDef->pSrcBuf, jStringLength, jChars, jStringLength, &jStringSize);
                        	jSource = (*env)->NewString(env, jChars, jStringLength);
                        	if (jChars != NULL && jChars != tmpBuf) {
                        		free(jChars);
                        	}
                    		jChars = NULL;
                        } else {
                        	jSource = (*env)->NewString(env, (jchar *)pDpFieldDataDef->pSrcBuf, jStringLength);
                        }
                        (*env)->SetObjectField(env, jObject, jFieldSourceFID, jSource);
                        (*env)->DeleteLocalRef(env, jSource);
                        break;
                }
            }
        }
        if (jErrorCode != 0 && pDpFieldDataDef->pDstBuf != NULL) {
        	/* We are going to use and abuse pDstBuf to contain the error message on error */
            // If wchar is 32-bit convert from platform UTF-32 to Java UTF-16, else use UTF-16 as is.
            jStringLength = (jsize) WCSLEN((ODPP_WCHAR *)pDpFieldDataDef->pDstBuf);
            if (cvtWchars) {
            	if (jStringLength + 1 <= sizeof(tmpBuf)/sizeof(jchar)){
            		jChars = tmpBuf;
            	} else {
            		jChars = (jchar *)malloc(sizeof(jchar) * (jStringLength + 1));
            	}
            	memset(jChars, 0, (jStringLength + 1) * sizeof(jchar));
                //jChars = (jchar *)malloc(sizeof(jchar) * (jStringLength + 1));
            	Convert32To16((ODPP_UINT *)pDpFieldDataDef->pDstBuf, jStringLength, jChars, jStringLength, &jStringSize);
            	jSource = (*env)->NewString(env, jChars, jStringLength);
            	if (jChars != NULL && jChars != tmpBuf) {
            		free(jChars);
            	}
        		jChars = NULL;
            } else {
            	jSource = (*env)->NewString(env, (jchar *)pDpFieldDataDef->pDstBuf, jStringLength);
            }
            (*env)->SetObjectField(env, jObject, jFieldErrorMessageFID, jSource);
            (*env)->DeleteLocalRef(env, jSource);
            free(pDpFieldDataDef->pDstBuf);
            pDpFieldDataDef->pDstBuf = (unsigned char *)NULL;
            (*env)->SetObjectField(env, jObject, jFieldDestinationFID, NULL);

        }
    }
	debugf("Field_toJava: Done.\n");
}

/*
 * Class:     com_ibm_nex_odpp_Field
 * Method:    free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_nex_odpp_Field_free(JNIEnv *env, jobject jObject)
{
    jlong jPointer;
    DP_FIELD_DATA_DEF *pDpFieldDataDef;
    RETURN_IF_NULL(jFieldPointerFID);
    debugf("Field_free: Freeing 'Field'...\n");
    jPointer = (*env)->GetLongField(env, jObject, jFieldPointerFID);
    debugf("Field_free: The pointer is 0x%16.16X\n", jPointer);
    if (jPointer != 0L) {
        pDpFieldDataDef = (DP_FIELD_DATA_DEF *)jPointer;
        CHECK_FIELD_DATA_DEF(pDpFieldDataDef)
        if (pDpFieldDataDef->pSrcBuf != NULL) {
            debugf("Field_free: Freeing source buffer...0x%16.16X\n", (unsigned long long)pDpFieldDataDef->pSrcBuf);
           	free(pDpFieldDataDef->pSrcBuf);
            pDpFieldDataDef->pSrcBuf = NULL;
        }
        if (pDpFieldDataDef->pDstBuf != NULL) {
            debugf("Field_free: Freeing destination buffer...0x%16.16X\n", (unsigned long long)pDpFieldDataDef->pDstBuf);
            free(pDpFieldDataDef->pDstBuf);
            pDpFieldDataDef->pDstBuf = NULL;
        }
        debugf("Field_free: Freeing 'DP_FIELD_DATA_DEF'...\n");
        free(pDpFieldDataDef);
        jPointer = (jlong) 0L;
        (*env)->SetLongField(env, jObject, jFieldPointerFID, jPointer);
        debugf("Field_free: Free of 'Field' done.\n");
    }
    debugf("Field_free: Done.\n");
}

/* Code below borrows heavily from ODPP documentation example on error handling */
static void checkForError(int svcToken, DP_ROWSET_DEF *pRowsetDef)
{
	RETVAL retVal;
	DP_ROW_DEF *pDpRowDef;
	DP_FIELD_DATA_DEF *pDpFieldDataDef;
	DP_ERR_MSG ErrMsg;
	DP_ECB *pEcb = (DP_ECB *)NULL;
	int iErrAreaLen; //Length of error control block storage area
	short sErrCnt = 0;
	short i = 0;
    jsize jStringLength;
    unsigned char foundErrors = FALSE;

	/* We need to put the messages on the heap, eventually. But temporaries can be stored on the stack */
	ODPP_WCHAR ErrMsgSrc[ODPP_MAX_ERR_MSG_SRC_LENGTH + 1];
	ODPP_WCHAR ErrMsgText[ODPP_MAX_ERR_MSG_TEXT_LENGTH + 1];
	ODPP_WCHAR ErrMsgBody[ODPP_MAX_ERR_MSG_BODY_LENGTH + 1];

	if (pRowsetDef == (DP_ROWSET_DEF *)NULL) {
		return;
	}
    CHECK_ROWSET_DEF(pRowsetDef)
	/* Iterate across rows in the row set */
	pDpRowDef = pRowsetDef->pRowDefine;
	debugf("checkForError, pDpRowDef=0x%16.16X\n", (unsigned long long)pDpRowDef);
	if (pDpRowDef == (DP_ROW_DEF *)NULL) {
		return;
	}
	do {
	    CHECK_ROW_DEF(pDpRowDef)
		/* Iterate across fields in the row */
		pDpFieldDataDef = pDpRowDef->pFldDataDefine;
		debugf("checkForError, pDpFieldDataDef=0x%16.16X\n", (unsigned long long)pDpFieldDataDef);
		if (pDpFieldDataDef == (DP_FIELD_DATA_DEF *)NULL) {
			continue;
		}
		do {
		    CHECK_FIELD_DATA_DEF(pDpFieldDataDef)
			if (pDpFieldDataDef->iErrorCode != 0) {
				foundErrors = TRUE;
				/* We have an error.  Now get the message */
				debugf("checkForError, pDpFieldDataDef=0x%16.16X has error code %d\n", (unsigned long long)pDpFieldDataDef, pDpFieldDataDef->iErrorCode);
				/* First, get the message count */
				memset(&ErrMsg, 0, sizeof(DP_ERR_MSG));

				//Initialize the DP_ERR_MSG structure
				INITIALIZE_ODPP_STRUCT(ErrMsg, DP_ERR_MSG)

				//Retrieve the number of errors that have occurred and the size of the ECB storage area
				retVal = Provider_GetErrorCount(svcToken, &sErrCnt, &iErrAreaLen);
				if(ODPPSUCCESS != retVal)
				{
					// Move on.
					continue;
				}
				if(sErrCnt <= 0 || iErrAreaLen == 0)
				{
					// Didn't find any errors. go on.
					continue;
				}
				debugf("\nTotal Error Count = %d for token %d\n", sErrCnt, svcToken);

				//Now set the required fields of DP_ERR_MSG

				//Language to use to return the error message. Currently not used and zero must be supplied.
				ErrMsg.iLanguage  = 0;

				//Allocate the ODPP error control block
				pEcb = (DP_ECB*)malloc(iErrAreaLen);
				if(NULL == pEcb)
				{
					return;  // Not much we can do here if malloc is failing
				}
				memset(pEcb, 0, iErrAreaLen);

				//Initialize the DP_ECB structure
				INITIALIZE_ODPP_STRUCT_PTR(pEcb, DP_ECB)
				//Set the subtype of DP_ERR_MSG to 'W' to retrieve formatted error message in wide character(Unicode) format
				ErrMsg.cSubType = 'W';

				//Allocate memory for error message block
				ErrMsg.EM.pWC = (DPEM_WC_SS *) malloc(sizeof(DPEM_WC_SS));
				if(NULL == ErrMsg.EM.pWC)
				{
					free(pEcb); // Malloc failed. No real point in continuing...
					pEcb = (DP_ECB*)NULL;
					continue;
				}
				memset(ErrMsg.EM.pWC, 0, sizeof(DPEM_WC_SS));

				//pMsgSrc will hold the source of the error message in wide character (Unicode) format
				ErrMsg.EM.pWC->pMsgSrc = &ErrMsgSrc[0];
				//Set the size of the buffer ErrMsgSrc, in bytes
				ErrMsg.EM.pWC->iMsgSrcBytes = sizeof(ErrMsgSrc);

				//pMsg will hold the error message text in wide character (Unicode) format
				ErrMsg.EM.pWC->pMsg = &ErrMsgText[0];
				//Set the size of the buffer ErrMsgText, in bytes
				ErrMsg.EM.pWC->iMsgBytes = sizeof(ErrMsgText);

				//pMsgBody will hold the error message description in details in wide character (Unicode) format
				ErrMsg.EM.pWC->pMsgBody = &ErrMsgBody[0];
				//Set the size of the buffer ErrMsgBody, in bytes
				ErrMsg.EM.pWC->iMsgBodyBytes = sizeof(ErrMsgBody);
				   //Loop through number of error count and print the formatted error messages
				for (i  =  0; i < sErrCnt; i++ )
				{
					//Call this function to get the oldest ECB
					retVal = Provider_GetError(svcToken, pEcb);

					debugf("\nRow Number: %d\n",(int)pEcb->iRowNum);
					debugf("Error Code: %d\t",pEcb->iErrCode);
					debugf("Token Count: %d\t\n",pEcb->sNumTokens);
				}
				//We now have the last error block.
				memset(ErrMsgSrc,'\0',sizeof(ErrMsgSrc));
				memset(ErrMsgText,'\0',sizeof(ErrMsgText));
				memset(ErrMsgBody,'\0',sizeof(ErrMsgBody));

				//Assign the error control block populated in call to Provider_GetError()
				ErrMsg.pECB = pEcb;

				//Retrieve the number of errors that have occurred and the size of the ECB storage area
				retVal = Provider_GetFormattedErrorMsg(&ErrMsg);
				if(retVal != ODPPSUCCESS)
				{
					debugf("Error Code not found, Err = %#x\n",retVal);
					if (ErrMsg.EM.pWC != NULL) {
						free(ErrMsg.EM.pWC);
						ErrMsg.EM.pWC = NULL;
					}
					if (pEcb != (DP_ECB*)NULL) {
						free(pEcb);
						pEcb = (DP_ECB*)NULL;
						ErrMsg.pECB = (DP_ECB*)NULL;
					}
					continue;
				}
				/* The sample code has this call to Provider_TermErrors here. It seems to be a VERY BAD IDEA */

			    debugf("Error Message Source:%S\n",ErrMsgSrc);
			    debugf("Error Message Text: %S \n",ErrMsgText);
			    debugf("Error Message Body: \n%S\n",ErrMsgBody);


				/* OK.  Conversion occurs later. Copy message to buffer */
	            jStringLength = (jsize) WCSLEN(ErrMsgText);
	            if (pDpFieldDataDef->pDstBuf != (unsigned char *)NULL) {
	            	debugf("WHOA NELLIE!!!\n");
	            } else {
					pDpFieldDataDef->pDstBuf = (unsigned char *)malloc(sizeof(ODPP_WCHAR) * (jStringLength + 1));
					memset(pDpFieldDataDef->pDstBuf, 0, sizeof(ODPP_WCHAR) * (jStringLength + 1));
					memcpy(pDpFieldDataDef->pDstBuf,ErrMsgText, jStringLength*sizeof(ODPP_WCHAR));
	            }
				free(ErrMsg.EM.pWC);
				free(pEcb);
				ErrMsg.EM.pWC = NULL;
				pEcb = (DP_ECB*)NULL;
				ErrMsg.pECB = (DP_ECB *)NULL;
			}
		} 	while ((pDpFieldDataDef = pDpFieldDataDef->pNext) != (DP_FIELD_DATA_DEF *)NULL);
	} while ((pDpRowDef = pDpRowDef->pNext) != (DP_ROW_DEF *)NULL);
	/* The call to Provider_TermErrors seems to cause memory HAVOC. NEVER, EVER call it */

}

static void throwODPPException(JNIEnv *env, const char *message, int code)
{
    jstring jMessage;
    jint jCode;
    jthrowable jException;
    jMessage = (*env)->NewStringUTF(env, message);
    jCode = code;
    jException = (*env)->NewObject(env, jODPPExceptionClass, jODPPExceptionInitMID, jMessage, jCode);
    (*env)->DeleteLocalRef(env, jMessage);
    (*env)->Throw(env, jException);
}

static void debugf(const char *format, ...)
{
    va_list ap;
    if (!debug) {
        return;
    }
    va_start(ap, format);
    vfprintf(stdout, format, ap);
    fflush(stdout);
    va_end(ap);
}

static void HexDump(const unsigned char *buffer, int size)
{
	int i;
	for (i=0; i < size; i ++) {
		if (i>0 && i%8==0) {
			(void)printf("\n");
		}
		(void)printf("0x%2.2X ", *(buffer+i));
	}
	(void)printf("\n");
}

/* NOTA BENE:  This function SHOULD become a part of some utility library.
 * SO, I'm not going to tailor it to its usage here.  Bobp 5/2/13
 */
//---------------------------------------------------------------------------
// Function:   ODPPParseParameters
//
// Desc:
// This function will take in a provider-syntax string and convert it out
// into a ODPP recognized DP_SVC_DEF structure.
// This resultant structure pSvcDef would need to be cleaned ones the
// work is complete which typically would be destructor of the class
// Use API ParmGen_Clean to clean this structure.
//
// Returns:
//    ODPPSuccess
//    Multiple error codes
//
// Parameters
//
// Author:  Ritesh Tandon, IBM ISL, 23/Jul/2012
//
//---------------------------------------------------------------------------
static
RETVAL ODPPParseParameters(char *pParamString,  //Provider Syntax string
                           int iParmStrBytes,   //Provider Syntax string bytes
                           int iCodePage,       //Provider Syntax string codepage
                           char cDBMSType,      //Provider Syntax string DBMSType
                           DP_SVC_DEF *pSvcDef, //ODPP required service definition
                                                //structure
                           char *pSvcId,        //Buffer for returning ODPP specific
                                                //service provider name
                           int iSvcIdBytes)     //bytes
{
   RETVAL retVal = ODPPFAILURE;
   RETVAL retVal1 = ODPPFAILURE;
   DP_PARAMS_DEF Params;
   DP_PARMGEN_DEF ParmGenDef;
   int iIdLen = 0;
   ODPP_BOOL bCleanPrmGen = FALSE;
   int i = 0;
   int iCnt = 0;

   debugf("ODPPParseParameters: Entry\n");
   /* Do basic validation */
   if ((pParamString == (char *)NULL) ||
	   (iParmStrBytes <= 0)           ||
	   (pSvcDef == (DP_SVC_DEF *)NULL)||
	   (pSvcId == (char *)NULL)       ||
	   (iSvcIdBytes <= 0)) {
	      debugf("ODPPParseParameters: Bad arguments\n");
	      retVal = ODPPFAILURE;
	      goto OnError;
   }
   memset(&Params, 0, sizeof(DP_PARAMS_DEF));
   INITIALIZE_ODPP_STRUCT(Params, DP_PARAMS_DEF)

   if (!parserInitialized) {
	   debugf("ODPPParseParameters: Calling Parser_Init()\n");
	   retVal = Parser_Init();
	   if(ODPPSUCCESS != retVal )
	   {
		  debugf("ODPPParseParameters: Parser_Init() fails with %d\n", retVal);
		  goto OnError;
	   }
	   parserInitialized = TRUE;
   } else {
	  debugf("ODPPParseParameters: parser already initialized\n");
   }


   //UTF8 or Latin string
   debugf("ODPPParseParameters: Setting up Params structure\n");
   Params.PS.pMC = (DPPS_MC_SS *)malloc(sizeof(DPPS_MC_SS));
   if(NULL  == Params.PS.pMC)
   {
      debugf("ODPPParseParameters: malloc() fails for Params.PS.pMC\n");
      retVal = ODPP_UDF_ERR_PRSR_PARAM_MC_MALLOC_FAILED;
      goto OnError;
   }
   memset(Params.PS.pMC, 0, sizeof(DPPS_MC_SS));
   Params.PS.pMC->pParamString = pParamString;
   Params.PS.pMC->iCodePage = iCodePage;
   Params.PS.pMC->cDBMSType = cDBMSType;
   Params.cParamStringType = 'M';
   Params.iParamStringBytes = iParmStrBytes;

   //First call Provider_Parse() method is called the parser returns the
   //number of Key-Value pairs required in pParamDef->iKeyCount, iKeyMaxBytes
   //and iValueMaxBytes have the max key and value bytes required which can
   //be used to allocate pKey and pValue*/
   debugf("ODPPParseParameters: Calling Parser_Service()\n");
   retVal = Parser_Service(&Params);
   if(retVal != ODPPSUCCESS)
   {
	  debugf("ODPPParseParameters: Parser_Service() fails with %d\n", retVal);
      goto OnError;
   }
   Params.iKeyValueArrSize = Params.iKeyCount;
   Params.pKeyValueArr = (DP_KVPAIR_DEF *)malloc(Params.iKeyValueArrSize
                                                   * sizeof(DP_KVPAIR_DEF));
   if(NULL == Params.pKeyValueArr)
   {
      debugf("ODPPParseParameters: Params.pKeyValueArr is NULL\n");
      retVal = ODPP_UDF_ERR_PRSR_KEYVAL_ARR_MALLOC_FAILED;
      goto OnError;
   }

   debugf("ODPPParseParameters: Initializing Parms.pKeyValueArr\n");
   memset(Params.pKeyValueArr, 0, Params.iKeyValueArrSize *
                                    sizeof(DP_KVPAIR_DEF));

   for(i = 0;i < Params.iKeyValueArrSize;i++)
   {
      INITIALIZE_ODPP_STRUCT((Params.pKeyValueArr[i]), DP_KVPAIR_DEF)
      (Params.pKeyValueArr[i]).iKeyBytes = Params.iKeyMaxBytes;
      (Params.pKeyValueArr[i]).pKey =
                  (ODPP_WCHAR *)malloc((Params.pKeyValueArr[i]).iKeyBytes);
      if(NULL == (Params.pKeyValueArr[i]).pKey)
      {
         retVal = ODPP_UDF_ERR_PRSR_KEY_MALLOC_FAILED;
         goto OnError;
      }

      memset((Params.pKeyValueArr[i]).pKey,
             0,
             ((Params.pKeyValueArr[i]).iKeyBytes));
      (Params.pKeyValueArr[i]).iValueBytes = Params.iValueMaxBytes;
      (Params.pKeyValueArr[i]).pValue =
                  (ODPP_WCHAR *)malloc((Params.pKeyValueArr[i]).iValueBytes);
      if(NULL == (Params.pKeyValueArr[i]).pKey)
      {
         retVal = ODPP_UDF_ERR_PRSR_VAL_MALLOC_FAILED;
         goto OnError;
      }
      memset((Params.pKeyValueArr[i]).pValue,
             0,
            ((Params.pKeyValueArr[i]).iValueBytes));
   }

   Params.cEnv = ENV_APP;

   //Invoke the ODPP Parser
   debugf("ODPPParseParameters: Invoking Parser_Service() again\n");
   retVal = Parser_Service(&Params);
   if(ODPPSUCCESS != retVal || Params.bHasError)
   {
      debugf("ODPPParseParameters: Parser_Service() invocation #2 fails with %d or Params.bHasError\n", retVal);
      goto OnError;
   }

   if (!parmGenInitialized) {
	   debugf("ODPPParseParameters: Invoking ParmGen_Init()\n");
	   retVal = ParmGen_Init();
	   if(ODPPSUCCESS != retVal)
	   {
		  debugf("ODPPParseParameters: ParmGen_Init() fails with %d\n", retVal);
		  goto OnError;
	   }
	   parmGenInitialized = TRUE;
	   //bTermParamGen = TRUE;
	   debugf("ODPPParseParameters: ParmGen_Init() succeeded. Initializing ParmGenDef.\n");
   } else {
	   debugf("ODPPParseParameters: ParmGen already initialized\n");
   }

   memset(pSvcDef, 0, sizeof(DP_SVC_DEF));
   memset(pSvcId, 0, iSvcIdBytes);
   memset(&ParmGenDef, 0, sizeof(DP_PARMGEN_DEF));

   INITIALIZE_ODPP_STRUCT(ParmGenDef, DP_PARMGEN_DEF)
   ParmGenDef.pParserParams = &Params;
   ParmGenDef.pSvcDef = pSvcDef;
   ParmGenDef.pSvcId = pSvcId;
   ParmGenDef.iSvcIdBufLen = iSvcIdBytes;
   ParmGenDef.pSvcIdLen = &iIdLen;

   debugf("ODPPParseParameters: Invoking ParmGen_Service()\n");
   retVal = ParmGen_Service(&ParmGenDef);
   if(ODPPSUCCESS != retVal)
   {
	  debugf("ODPPParseParameters: ParmGen_Service() fails with %d\n", retVal);
      bCleanPrmGen = TRUE;
      goto OnError;
   }

   bCleanPrmGen = FALSE;

   retVal = ODPPSUCCESS;

OnError:
   for(iCnt = 0; iCnt < Params.iKeyValueArrSize; iCnt++)
   {
      if(NULL != (Params.pKeyValueArr[iCnt]).pKey) {
         free((Params.pKeyValueArr[iCnt]).pKey);
         (Params.pKeyValueArr[iCnt]).pKey = NULL;
      }
      if(NULL != (Params.pKeyValueArr[iCnt]).pValue) {
         free((Params.pKeyValueArr[iCnt]).pValue);
         (Params.pKeyValueArr[iCnt]).pValue = NULL;
      }
   }
   switch(Params.cParamStringType)
   {
   case 'M':
      if(NULL != Params.PS.pMC) {
         free(Params.PS.pMC);
         Params.PS.pMC = NULL;
      }
      break;
   default:
      if(NULL != Params.PS.pWC) {
         free(Params.PS.pWC);
         Params.PS.pWC = NULL;
      }
      break;
   }


   if(TRUE == bCleanPrmGen)
   {
 	  debugf("ODPPParseParameters: Invoking ParmGen_Clean\n");
      retVal1 = ParmGen_Clean(pSvcDef);
      if(ODPPSUCCESS != retVal1)
      {
    	  debugf("ODPPParseParameters: ParmGen_Clean() fails with %d\n", retVal1);
         //trace it
      }
      pSvcDef = NULL; // Not really necessary, since we aren't using it again. But... good habit
   }

   debugf("ODPPParseParameters: Returning %d\n", retVal);
   return retVal;
}
